"""
aLIGO Guardian python library
"""

# tell guardian to not write bytecode (.pyc) files.  This is meant to
# circumvent the creation of .pyc files for usercode, which tend to
# cause confusion and clutter much more than providing any system
# speedup.
import sys
sys.dont_write_bytecode = True

from ._version import __version__
from .state import GuardState, GuardStateDecorator
from .system import GuardSystem

try:
    from .manager import Node, NodeManager
except ImportError:
    pass



import sys
import time
import signal
import logging
import inspect
import traceback
import multiprocessing
if sys.version_info.major > 2:
    import builtins
else:
    import __builtin__ as builtins
try:
    from setproctitle import setproctitle
except ImportError:
    def setproctitle(*args):
        pass

from ezca import __version__ as ezca__version__
from ezca import Ezca, EzcaError, EzcaConnectError

from .db import WORKER_STATUS
from .log import StateLogger
from .shmem import SharedMemEnum, SharedMemString
from .const import USERMSG_STRING_LENGTH, USERMSG_COUNT, STATE_NAME_LENGTH, SPM_DIFF_COUNT
from .manager import Node, NodeManager

############################################################

WORKER_CMD = ('MAIN', 'RUN', 'NOOP')

############################################################

class WorkerLogger(logging.LoggerAdapter):
    def __init__(self, logger):
        super(WorkerLogger, self).__init__(logger, {})

    def process(self, msg, kwargs):
        msg = 'W: %s' % msg
        return (msg, kwargs)

############################################################
# The worker is what executes all user code.  It is run in a separate
# process so that the main process can terminate it as needed, given
# that the user code may take arbitrarily long to execute.

class Worker(multiprocessing.Process):
    """Guardian daemon worker subprocess.

    """
    def __init__(self, system, logger, snapshot_file=None):
        super(Worker, self).__init__()

        self._system = system
        self._log = WorkerLogger(logger)
        self._statelogger = StateLogger(logger)

        self._userlog_message = ''
        self._usernotify_messages = []

        # shared memory objects
        self._shmem = {}
        self._shmem['CMD'] = SharedMemEnum(WORKER_CMD)
        self._shmem['STATE'] = SharedMemString(size=STATE_NAME_LENGTH)
        self._shmem['EXECSTART'] = multiprocessing.Value('d', lock=True)
        self._shmem['EXECSTOP'] = multiprocessing.Value('d', lock=True)
        self._shmem['STATUS'] = SharedMemEnum(WORKER_STATUS)
        self._shmem['LOGLEVEL'] = SharedMemEnum(('INFO', 'DEBUG', 'LOOP'))
        self._shmem['JUMP'] = SharedMemString(size=STATE_NAME_LENGTH)
        for i in range(USERMSG_COUNT):
            self._shmem['USERMSG'+str(i)] = SharedMemString(size=USERMSG_STRING_LENGTH)
        self._shmem['NUSERMSGS'] = multiprocessing.Value('H', lock=True)
        self._shmem['PV_TOTAL'] = multiprocessing.Value('H', lock=True)
        self._shmem['SPM_TOTAL'] = multiprocessing.Value('H', lock=True)
        self._shmem['SPM_CHANGED'] = multiprocessing.Value('H', lock=True)
        for i in range(SPM_DIFF_COUNT):
            self._shmem['SPM_DIFF'+str(i)] = SharedMemString(size=STATE_NAME_LENGTH)
            self._shmem['SPM_DIFF'+str(i)+'_S'] = SharedMemString(size=USERMSG_STRING_LENGTH)
            self._shmem['SPM_DIFF'+str(i)+'_C'] = SharedMemString(size=USERMSG_STRING_LENGTH)
            self._shmem['SPM_DIFF'+str(i)+'_D'] = SharedMemString(size=USERMSG_STRING_LENGTH)
        self._shmem['SUBNODES_TOTAL'] = multiprocessing.Value('H', lock=True)
        self._shmem['SUBNODES_NOT_OK'] = multiprocessing.Value('H', lock=True)

        # load request event
        self.load_request = multiprocessing.Event()

        # PV snapshots
        self.snapshot_request = multiprocessing.Event()
        self.snapshot_file = snapshot_file

        # run lock. it uses an reentrant lock (RLock) by default, but
        # we don't want anyone to be able to claim the lock more then
        # once, so we pass it a regular Lock object instead.
        self._run_lock = multiprocessing.Condition(multiprocessing.Lock())

        # ready event, used to notify daemon that worker is ready to
        # start processing.
        self._ready_event = multiprocessing.Event()

        # make this process a daemon, so it's necessarily killed when
        # the parent process exits
        self.daemon = True

        self._log.info("initialized")

    ########################################
    # shared memory access

    def __getitem__(self, shmem):
        return self._shmem[shmem].value

    def __setitem__(self, shmem, value):
        self._shmem[shmem].value = value

    ########################################

    def command(self, state, cmd):
        """Set command for worker."""
        self._log.log(5, "command received: %s %s" % (state, cmd))
        self['STATE'] = state
        self['CMD'] = cmd
        self._run_lock.notify()

    def acquire(self):
        """Acquire the worker run lock (non-blocking)."""
        return self._run_lock.acquire(False)

    def release(self):
        """Release the worker run lock."""
        self._run_lock.release()

    ########################################

    # user notify and log functions
    def _userlog(self, message=''):
        if message == self._userlog_message:
            return
        self._statelogger.warning(message)
        self._userlog_message = message

    def _usernotify(self, message=None):
        # if no message, clear the message cache
        if not message:
            if self._usernotify_messages:
                self._statelogger.warning('USERMSG cleared')
            self._usernotify_messages = []
        else:
            # append message to cache if it's not already there
            if message not in self._usernotify_messages:
                self._usernotify_messages.append(message)
            if message not in self._usernotify_messages_last:
                self._statelogger.warning('USERMSG %d: %s' % (self._usernotify_messages.index(message), message))
            # push message if it's not already in the stack
            if message[:USERMSG_STRING_LENGTH] not in [self['USERMSG'+str(i)].strip('\0') for i in range(USERMSG_COUNT)]:
                ind = len(self._usernotify_messages) - 1
                if ind >= 0 and ind < 10:
                    self['USERMSG'+str(ind)] = self._usernotify_messages[ind]

    ########################################

    def spm_diff_set(self, i, chan='', S='', C='', D=''):
        if i >= SPM_DIFF_COUNT:
            return
        sc = 'SPM_DIFF'+str(i)
        self[sc] = chan
        self[sc+'_S'] = S
        self[sc+'_C'] = C
        self[sc+'_D'] = D

    ########################################

    def start(self):
        """Start worker process."""
        super(Worker, self).start()
        while not self._ready_event.is_set():
            if self.exitcode is not None:
                raise SystemExit(self.exitcode)

    def run(self):
        setproctitle('guardian-worker {} {}'.format(self._system.name, self._system.path))

        # reset the sigterm handler in the worker back to the default.
        # a custom sigterm handler was installed in main which
        # properly terminates the worker.  the worker needs to reset
        # to the default handler to be able to handle the terminate
        # signals from the main process
        # FIXME: this shouldn't be necessary since daemon=True should
        # take care of this
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        # FIXME: the following log/state/ezca initialization seems
        # more convoluted than it should be.  it's basically all about
        # recording state in all ezca and state log calls.  we want
        # ezca initialized once for all states, but we want it's log
        # to have the state info.  so we have to create a logger
        # object first, then pass it to ezca and state objects at
        # initialization.  then we update the state in the logger for
        # each new state, so both ezca and state log have the info
        # when they log.
        #
        # A better way might be to actually initialize the EPICS
        # device *outside* of ezca, and initialize the ezca object
        # with the device and log objects during every state.

        # the following variables are made global for state code by
        # adding it to builtins.  this allows user code to access them
        # from anywhere, without having to pass around the objects
        # from the state.

        # initialize ezca EPICS channel access.
        # EPICS devices need to be initialized after forking, since
        # the EPICS device appears to be tied to the process ID.
        # FIXME: can the object somehow be passed from the parent?
        # FIXME: another alternative would be to setup an Ezca proxy
        # server in a separate process that never dies
        builtins.ezca = Ezca(prefix=self._system.ca_prefix,
                             logger=self._statelogger)
        self._log.info("EZCA v{}".format(ezca__version__))
        self._log.info("EZCA CA prefix: {}".format(ezca.prefix))
        if self._system.ca_channels:
            self._log.info("initializing setpoints...")
            ezca.init_setpoints(self._system.ca_channels)

        builtins.notify = self._usernotify
        builtins.log = self._userlog

        ########################################

        # the current state object itself
        state = None

        # the condition lock must first be acquired before it can be
        # released by the run_lock.wait() in the begining of the while
        # loop that follows
        self._run_lock.acquire()

        # notify that the worker has fully started
        self._ready_event.set()

        self._log.info("ready")

        ########################################
        ########################################
        while True:

            # clean up USERMSGs and update count.  if there were no
            # notifications this cycle, this should clear all usermsgs
            self['NUSERMSGS'] = len(self._usernotify_messages)
            for i in range(USERMSG_COUNT):
                try:
                    self['USERMSG'+str(i)] = self._usernotify_messages[i]
                except IndexError:
                    self['USERMSG'+str(i)] = ''
            self._usernotify_messages_last = self._usernotify_messages
            self._usernotify_messages = []

            # note command stop time, which happens when we release
            # the lock
            if self['STATUS'] != 'INIT':
                self['EXECSTOP'] = time.time()

            ##############################
            ##############################
            try:
                self._run_lock.wait()
            except KeyboardInterrupt:
                raise SystemExit()
            ##############################
            ##############################

            # note command start time
            self['EXECSTART'] = time.time()

            self._log.logger.setLevel(self['LOGLEVEL'])

            ##############################

            # clear all setpoint diff messages on reload
            if self.load_request.is_set():
                self['SPM_CHANGED'] = 0
                for i in range(SPM_DIFF_COUNT):
                    self.spm_diff_set(i)

            ##############################
            if self['CMD'] == 'MAIN':

                ##########
                # initialize the state object
                # we always want a freshly initialized state oject, so
                # there is no chance of anything being leftover from
                # a previous state visit.

                # start by cleaning up and destroying any previous
                # state objects
                # FIXME: it might be best to allow the main process to
                # specifically initiate cleanup as needed.  leaving it
                # until initialization of the next time could leave
                # callbacks or threads in place that might get
                # executed at inopportune times.
                if state:
                    # execute the state destroy method to terminate
                    # any internal threads
                    state.destroy()
                    # FIXME: does setting state to None actually cause
                    # the object to be fully garbage collected?
                    state = None

                if self.load_request.is_set():
                    self._log.info('RELOADING @ %s.main' % self['STATE'])
                    # reload system
                    self._system.load()
                    # initialize setpoints if specified
                    if self._system.ca_channels:
                        self._log.info("re-initializing setpoints...")
                        try:
                            ezca.init_setpoints(self._system.ca_channels)
                        except (EzcaError, EzcaConnectError):
                            self._log.error(traceback.format_exc().splitlines()[-1])
                            self['STATUS'] = 'ERROR'
                            continue
                    # clear the load queue
                    self.load_request.clear()
                    self._log.debug('RELOAD @ %s.main' % self['STATE'])

                self._log.log(5, "%s.enter:", self['STATE'])
                self._statelogger.set_state(self['STATE'])
                self._statelogger.set_func('enter')
                state = self._system[self['STATE']](logfunc=self._userlog)
                self._userlog(' ')
                self._log.log(5, "%s.enter.", self['STATE'])

                ##########

                statefunc = state.main

            ##############################
            elif self['CMD'] == 'RUN':

                ##########
                if self.load_request.is_set():
                    self._log.info('RELOADING @ %s.run' % self['STATE'])
                    # get a list of members in the current state object
                    objects = inspect.getmembers(state)
                    # reload system
                    self._system.load()
                    # re-initialize the state from the newly loaded
                    # system
                    state = self._system[self['STATE']](logfunc=self._userlog)
                    # update the attributes in the new state with the
                    # objects collected from the old state
                    for obj in objects:
                        # skip build-in objects
                        if obj[0][:2] == '__':
                            continue
                        # skip primary methods themselves.
                        if obj[0] in ['main', 'run', 'destroy']:
                            continue
                        try:
                            setattr(state, obj[0], obj[1])
                            self._log_debug("preserving: %s" % obj)
                        except:
                            # pass on any attributes that can't be set
                            pass
                    # initialize setpoints if specified
                    if self._system.ca_channels:
                        self._log.info("re-initializing setpoints...")
                        try:
                            ezca.init_setpoints(self._system.ca_channels)
                        except (EzcaError, EzcaConnectError):
                            self._log.error(traceback.format_exc().splitlines()[-1])
                            self['STATUS'] = 'ERROR'
                            continue
                    # clear the load queue
                    self.load_request.clear()
                    self._log.debug('RELOAD @ %s.run' % self['STATE'])
                ##########

                statefunc = state.run

            ##############################
            elif self['CMD'] == 'NOOP':

                # NOTE: NOOP should NOT change the STATUS, since that
                # can change the state while in pause.  is there a
                # cleaner way?
                statefunc = lambda: None

            ##############################
            ##############################

            # NOTE: shmem numbers are initialized to zero.  but that
            # number is probably bogus after initialization. all the
            # following checks are done before setting the worker
            # status to COMMAND, so that the bogus values are only
            # there during INIT.  the validity of these values should
            # therefore only be assumed when the status is *not* INIT.

            self['PV_TOTAL'] = len(ezca.pvs)
            self['SPM_TOTAL'] = len(ezca.setpoints)

            # SPM snapshots
            if self.snapshot_request.is_set():
                self._log.info("PVs:")
                for channel, pv in sorted(ezca.pvs.iteritems()):
                    self._log.info("  %s = %s" % (pv.pvname, pv.char_value))
                self._log.info("SPMs:")
                for channel, sp in sorted(ezca.setpoints.iteritems()):
                    self._log.info("  %s = %s" % (sp[0], sp[1]))
                self._log.info("%d PVs, %d SPMs" % (self['PV_TOTAL'], self['SPM_TOTAL']))
                if self.snapshot_file:
                    ezca.setpoint_snap(self.snapshot_file)
                    self._log.info("SPM snapshot: %s" % (self.snapshot_file))
                self.snapshot_request.clear()

            SPM_index = 0

            # CHECK CONNECTIONS
            dead_channels = ezca.check_connections()
            if dead_channels:
                self._usernotify("CONNECTION ERRORS. see SPM DIFFS for dead channels")
                for chan in dead_channels:
                    self.spm_diff_set(SPM_index, chan, '', 'DEAD', '')
                    SPM_index += 1

            # CHECK SETPOINTS
            if self._system.ca_monitor:
                changed = ezca.check_setpoints()
                if changed:
                    if self._system.ca_monitor_notify:
                        self._usernotify("SETPOINT CHANGES. see SPM DIFFS for differences")
                    for change in changed:
                        if change[0] in dead_channels:
                            continue
                        self.spm_diff_set(SPM_index, *change)
                        SPM_index += 1

            # clear remaining diffs
            if SPM_index < SPM_DIFF_COUNT:
                for i in range(SPM_index, SPM_DIFF_COUNT):
                    self.spm_diff_set(i)

            # note total number of differences
            self['SPM_CHANGED'] = SPM_index

            # if there were any dead channels don't proceed with code
            # execution
            if dead_channels:
                self['STATUS'] = 'CERROR'
                continue

            # CHECK MANAGER SUBORDINATES
            try:
                subnodes_total = set()
                subnodes_not_ok = set()
                for mngr in self._system.node_managers:
                    # make sure all nodes are initialized (node.init() is idempotent)
                    mngr.init()
                    # count subordinates nodes
                    # FIXME: clean up the interfaces to make this simpler
                    if isinstance(mngr, Node):
                        subnodes_total.add(mngr.name)
                        if not mngr.OK:
                            subnodes_not_ok.add(mngr.name)
                    elif isinstance(mngr, NodeManager):
                        subnodes_total |= set(mngr.nodes.keys())
                        subnodes_not_ok |= mngr.not_ok()
                    # FIXME: should we be running check_fault() here?
                self['SUBNODES_TOTAL'] = len(subnodes_total)
                self['SUBNODES_NOT_OK'] = len(subnodes_not_ok)
            except EzcaConnectError as e:
                self._usernotify("EZCA CONNECTION ERROR: %s" % e)
                self['STATUS'] = 'CERROR'
                continue

            ##############################
            ##############################

            # FIXME: continue here so that NOOP doesn't change STATUS.
            # is there a cleaner way to do this?
            if self['CMD'] == 'NOOP':
                continue

            ##############################

            self['STATUS'] = 'COMMAND'

            cmd = self['CMD'].lower()
            self._log.log(5, "%s.%s:", self['STATE'], cmd)
            self._statelogger.set_func(cmd)

            try:

                ###########
                # USER CODE
                retval = statefunc()
                # USER CODE
                ###########

            except EzcaConnectError as e:
                self._usernotify("EZCA CONNECTION ERROR: %s" % e)
                self['STATUS'] = 'CERROR'
                continue
            except KeyboardInterrupt:
                raise SystemExit
            except:
                # HACK: we should just compare the exception directly,
                # but we don't want to import all of cdsutils here
                # just to do that.  Maybe we should?
                e = traceback.format_exc(0).splitlines()[-1]
                if 'NDSConnectError' in e:
                    self._usernotify("NDS CONNECTION ERROR: %s" % e)
                    self['STATUS'] = 'CERROR'
                else:
                    self._log.error(traceback.format_exc())
                    self._usernotify("USER CODE ERROR (see log)")
                    self['STATUS'] = 'ERROR'
                continue

            self._log.log(5, "%s.%s.", self['STATE'], cmd)

            # string return values are jump state names
            if isinstance(retval, str):
                if retval in self._system:
                    self['JUMP'] = retval
                    self._log.debug("JUMP: %s", retval)
                    self['STATUS'] = 'JUMP'
                else:
                    self._log.error("ERROR: state returned unknown jump state: %s" % retval)
                    self['STATUS'] = 'ERROR'
            elif retval in [None, False]:
                self['STATUS'] = 'CONTINUE'
            elif retval is True:
                self['STATUS'] = 'DONE'
            else:
                self._log.error("state %s returned unexpected type: %s" % (self['STATE'], type(retval)))
                self['STATUS'] = 'ERROR'



import os

# needed variables
ENV = dict(
    IFO=None,
    GUARD_CHANFILE=None,
)

# load from conf file
CONFFILE = os.getenv('GUARD_CONFFILE', '/etc/guardian/local-env')
if os.path.exists(CONFFILE):
    with open(CONFFILE, 'r') as f:
        for line in f:
            if not line.strip():
                continue
            if line.strip()[0] == '#':
                continue
            try:
                name, value = line.strip().split('=', 1)
            except ValueError:
                raise SystemExit("invalid variable definition: {}".format(line))
            if name != name.upper():
                raise SystemExit("invalid variable name (not upper): {}".format(name))
            if name in ENV:
                ENV[name] = value

# load from environment
for name, value in list(ENV.items()):
    ENV[name] = os.getenv(name, value)



import os
import git
import shutil
import sys
if sys.version_info.major > 2:
    int_type = int
else:
    int_type = (int, long)

##################################################

def gen_gitpath(path):
    return os.path.relpath(path, '/')

def git_hex2int(hexsha, digits=7):
    """Convert git hexidecimal commit id into an integer.

    Take the first 'digits' of the hex and return an int.  The default
    'digits' is 7, which is a "standard" short git hex length that
    should be relatively unique, and is 28 bits, which fortunately
    fits into a standard EPICS 32-bit signed int.

    """
    return int(hexsha[:digits], 16)

def git_int2hex(intsha, digits=7):
    """Convert integer into hex.

    The 'intsha' is converted to a hex string.  If hex is less than
    'digits' in length (default 7), the string is zero padded up to
    'digits'.

    """
    return hex(intsha)[2:].zfill(digits)

##################################################

class SystemArchiveGitError(Exception):
    def __init__(self, e):
        error = 'Not a valid object name'
        if error in str(e):
            super(SystemArchiveGitError, self).__init__(error)

class SystemArchiveGit(object):
    def __init__(self, root, rw=False):
        self.root = os.path.abspath(root)
        self.repo = None
        # config = repo.config_writer()
        # config.set_value("user", "email", "")
        # config.set_value("user", "name", "Guardian Daemon")

    def __enter__(self):
        self.repo = git.Repo.init(self.root)
        return self

    def __exit__(self, *args):
        del self.repo
        self.repo = None

    def commit(self, system, message):
        """Commit system code base to archive git repo.

        Arguments are system to be archived and a commit message.

        Returns SHA1 hex commit id of commit.

        """
        if not self.repo:
            raise SystemArchiveGitError("repo not open, use context manager")
        repo = self.repo
        root = repo.working_dir

        # clean up the working tree, in case a previous commit was
        # aborted or hung.
        # reset the index and working tree to the HEAD.
        try:
            repo.head.reset(index=True, working_tree=True)
        except:
            pass
        # remove any stale index locks
        # FIXME: this should not be necessary, and undermines the
        # purpose of the lock
        try:
            os.remove(os.path.join(root, '.git', 'index.lock'))
        except:
            pass

        # completely remove all files from the index, so we commit
        # only a clean set of files (to handle code removes)
        if repo.head.is_valid():
            for entry in repo.head.commit.tree.traverse():
                fullpath = os.path.join(root, entry.path)
                if os.path.isfile(fullpath) or os.path.islink(fullpath):
                    repo.index.remove([entry.path])
        # FIXME: should remove *all* files, not just those known to
        # index, e.g.:
        # for root, dirs, files in os.walk(root):
        #     dirs[:] = [d for d in dirs if d != '.git']
        #     ...

        # copy and add all usercode files
        for path in [system.path] + system.usercode:
            # store all files in the repo at their current global
            # system path location, with the git working directory as
            # the filesystem root
            gitpath = gen_gitpath(path)
            outpath = os.path.join(root, gitpath)
            outdir = os.path.dirname(outpath)
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            shutil.copyfile(path, outpath)
            repo.index.add([gitpath])

        # add the guardian path as a PATH description in the
        # GUARD_MODULE_PATH file
        guardpath_file = os.path.join(root, 'GUARD_MODULE_PATH')
        with open(guardpath_file, 'w') as f:
            path = ':'.join([path[1:] for path in system.guardpath])
            f.write(path)
            f.write('\n')
        repo.index.add([guardpath_file])

        # if there are no differences detected, return None
        if repo.head.is_valid() and not repo.index.diff(repo.head.commit):
            return

        # commit the index
        repo.index.commit(message)

        # return the current commit hex SHA1
        return self.get_hexsha()

    def get_hexsha(self):
        return git.Repo.init(self.root).head.commit.hexsha

    def get_intsha(self, digits=7):
        return git_hex2int(self.get_hexsha(), digits=digits)

    def _resolve_id(self, id):
        if isinstance(id, int_type):
            return git_int2hex(id)
        else:
            return id

    def get_guardpath(self):
        with open(os.path.join(self.root, 'GUARD_MODULE_PATH')) as f:
            data = f.read()
        return [os.path.join(self.root, path) for path in data.split(':')]



import functools

from .state import GuardStateDecorator


class NodeError(Exception):
    pass


def node_checker(node_manager, fail_return=None):
    """Return GuardStateDecorator for checking fault status of Nodes.

    `node_manager` is a Node or NodeManager object with a
    check_fault() method.  Returns a GuardStateDecorator with it's
    pre_exec method set to be the check_fault method.  The
    "fail_return" option should specify an alternate return value for
    the decorated state method in case the check fails (i.e. a jump
    state name) (default None).

    """
    # we want to return a GuardStateDecorator here, where the
    # pre_exec function is the NodeManager.check method.  since
    # "self" here is the NodeManager, we tell pre_exec to just
    # ignore it's argument, which would have been the
    # GuardStateDecorator self.
    class checker(GuardStateDecorator):
        def pre_exec(__):
            if node_manager.check_fault():
                return fail_return
    return checker


# FIXME: handle rejected requests
# FIXME: check for dead node

class Node(object):
    """Manager interface to a single Guardian node.

    >>> SUS_ETMX = Node('SUS_ETMX')  # create the node object
    >>> SUS_ETMX.init()              # initialize (handled automatically in daemon)
    >>> SUS_ETMX.set_managed()       # set node to be in MANAGED mode
    >>> SUS_ETMX.set_request('DAMPED') # request DAMPED state from node
    >>> SUS_ETMX.arrived             # True if node arrived at requested state
    >>> SUS_ETMX.check_fault()       # Check for management-related "faults" in the Node
    >>> SUS_ETMX.release()           # release node from management

    """

    # node attributes
    attrs=('OP',
           'MODE',
           'MANAGER',
           'REQUEST',
           'STATE_S',
           'TARGET_S',
           'REQUEST_S',
           'STATUS',
           'STALLED',
           'ERROR',
           'NOTIFICATION',
           )

    def __init__(self, name):
        self.__name = name
        self.__prefix = ':GRD-{name}_'.format(name=self.name)
        # cache that we have set node to be managed
        self.__managed = False
        # cache of requested state
        self.__request = None
        # initialization state
        self.__initialized = False

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.__name)

    def __str__(self):
        if self.__initialized:
            s = 'initialized'
        else:
            s = 'uninitialized'
        return "<%s '%s', %s>" % (self.__class__.__name__, self.__name, s)

    @property
    def name(self):
        """Node name"""
        return self.__name

    def __get(self, attr):
        """get attribute of node."""
        return ezca.read(self.__prefix+attr, as_string=True)

    def __put(self, attr, value):
        """set attribute of node."""
        assert attr in ['MODE', 'MANAGER', 'REQUEST'], "MANAGER, MODE and REQUEST are the only settable node attributes."
        return ezca.write(self.__prefix+attr, value)

    ########################################

    def init(self):
        """Initialize the node.

        Under normal circumstances, i.e. in a running guardian daemon,
        node initialization is handled automatically.  This function
        therefore does not need to be executed in user code.

        """
        # FIXME: could elliminate this enitrely by just using the rote
        # ezca object
        if self.__initialized:
            return

        # initialize node PVs
        for attr in self.attrs:
            ezca.connect(self.__prefix+attr)

        # if the node is currently set to be managed by me
        if self.i_manage:

            # update the internal manage indicator
            self.__managed = True

            # set the internal request indicator based on the current
            # request (assuming it was set by me)
            if not self.__request:
                self.__request = self.REQUEST

        self.__initialized = True

    def __eq__(self, state):
        """True if node state string equals string."""
        return self.state == state

    def __ne__(self, state):
        """True if node state string does not equal string."""
        return not self.__eq__(state)

    @property
    def OP(self):
        """node OP"""
        return self.__get('OP')

    @property
    def MODE(self):
        """node MODE"""
        return self.__get('MODE')

    @property
    def managed(self):
        """True if node is MANAGED"""
        return self.MODE == 'MANAGED'

    @property
    def MANAGER(self):
        """MANAGER string of node"""
        return self.__get('MANAGER')

    def set_managed(self):
        """Set node to be managed by this manager."""
        # note that we have set node to be managed
        self.__managed = True
        self.__put('MANAGER', _SYSTEM_)

    def release(self):
        """Release node from management by this manager (MODE=>AUTO)."""
        log("Releasing node from management: %s" % self.name)
        self.__put('MODE', 'AUTO')
        self.__managed = False

    @property
    def i_manage(self):
        """True if node is being managed by this system"""
        return self.MANAGER == _SYSTEM_

    @property
    def ERROR(self):
        """True if node in ERROR."""
        return eval(self.__get('ERROR'))

    @property
    def NOTIFICATION(self):
        """True if node NOTIFICATION present."""
        return eval(self.__get('NOTIFICATION'))

    @property
    def OK(self):
        """Current OK status of node."""
        return eval(self.__get('OK'))

    @property
    def REQUEST(self):
        """Current REQUEST state of node."""
        return self.__get('REQUEST_S')
    request = REQUEST

    def set_request(self, state):
        """Set REQUEST state for node.

        """
        assert state != None
        self.__put('REQUEST', state)
        # HACK: We explicitly get the value without relying on the
        # monitor (use_monitor=False) to make sure we have the
        # post-update value.  If it doesn't match what we put then the
        # value was invalid.  Should be a better way to do this.
        pv = ezca.connect(self.__prefix+'REQUEST')
        if pv.get(use_monitor=False) != state:
            raise NodeError("Invalid REQUEST: {}".format(state))
        self.__request = state
        # FIXME: set call back to record when subordinate makes it to
        # request, for benefit of arrived.

    @property
    def STATE(self):
        """Current STATE of node."""
        return self.__get('STATE_S')
    state = STATE

    @property
    def TARGET(self):
        """Current TARGET state of node."""
        return self.__get('TARGET_S')

    @property
    def arrived(self):
        """True if node STATE equals the last manager-requested state.

        NOTE: This will be False if STATE == REQUEST but REQUEST was
        not last set by this Node manager object.  This prevents false
        positives in the case that the REQUEST has been changed out of
        band.

        """
        # FIXME: Maybe this should be a latching value?  It could be
        # reset when the request is set, and set True once it arrives,
        # and then not change when it looses the requested state.
        # Would need some sort of request callback for this.

        # assume true if no request has ever been issued for this
        # node.
        if not self.__request:
            return True
        else:
            return self.STATE == self.__request

    @property
    def STATUS(self):
        """Current STATUS of node."""
        return self.__get('STATUS')

    @property
    def done(self):
        """True if STATUS is DONE.

        A state is DONE if it is the requested state and the state
        method has returned True.

        """
        return self.STATUS == 'DONE'

    @property
    def completed(self):
        """True is node has arrived at the request state, and state is done.

        """
        return self.arrived and self.done

    @property
    def STALLED(self):
        """True if the node has stalled in the current state.

        This is true when STATE == TARGET != REQUEST, which is
        typically the result of a jump transition while in managed
        mode.

        """
        return eval(self.__get('STALLED'))

    def revive(self):
        """Re-request last requested state.

        The last requested state in this case is the one requested
        from this Node object.

        Useful for reviving stalled nodes, basically counteracting the
        stalling that is the effect of a jump transition while being
        in MANAGED mode.  See the 'STALLED' property.

        """
        if self.__request and self.STALLED:
            self.set_request(self.__request)

    def check_fault(self):
        """Return fault status of node.

        Runs a series of checks on the "management status" of the
        node, and returns True if any of the following checks fail:
         * node still alive and running
         * node does not show ERROR status
         * REQUEST hasn't deviated from last set value
         * if node had been set MANAGED, it is still set, and MANAGER
           hasn't changed
         * node has no notifications (failure does not produce fault)

        Any failure of the above also produces a NOTIFICATION message.

        """
        fault = False

        def snotify(msg):
            notify("{}: {}".format(self.name, msg))

        if self.NOTIFICATION:
            snotify("has notification")

        if self.ERROR:
            snotify("ERROR!")
            fault = True

        if self.OP != 'EXEC':
            snotify("not EXEC")
            fault = True

        if self.__request:
            request = self.REQUEST
            if request != self.__request:
                snotify("REQUEST CHANGED (was: %s, now: %s)" % (self.__request, request))
                fault = True

        if self.__managed:
            if not self.managed:
                snotify("NOT IN MANAGED MODE")
                fault = True

            # if we're not listed as manager
            elif not self.i_manage:
                snotify("STOLEN (by: %s)" % self.MANAGER)
                fault = True

        return fault

    @functools.wraps(node_checker)
    def checker(self, fail_return=None):
        return node_checker(self, fail_return=fail_return)


class NodeManager(object):
    """Manager interface to a set of subordinate Guardian nodes.

    This should be instantiated with a list of node names to be
    managed.  Node objects are instantiated for each node.

    >>> nodes = NodeManager(['SUS_ITMX','SUS_ETMX'])
    >>> nodes.init()                   # initialize (handled automatically in daemon)
    >>> nodes.set_managed()            # set all nodes to be in MANAGED mode
    >>> nodes['SUS_ETMX'] = 'ALIGNED'  # request state of node
    >>> nodes['SUS_ITMX'] = 'ALIGNED'  # request state of node
    >>> nodes.arrived                  # True if all nodes have arrived at their
                                       # requested states
    >>> nodes.check_fault()            # Check for management-related "faults" in all nodes

    """
    def __init__(self, nodes):
        self.nodes = {}
        for node in nodes:
            self.nodes[node] = Node(node)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.nodes.keys())

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.nodes.keys())

    def __getitem__(self, node):
        """Retrieve Node object for named node."""
        return self.nodes[node]

    def __contains__(self, node):
        """True if manager contains node."""
        if isinstance(node, Node):
            return node.name in self.nodes
        else:
            return node in self.nodes

    def __setitem__(self, node, state):
        """Request state for named node."""
        try:
            self.nodes[node].set_request(state)
        except NodeError as e:
            raise NodeError("%s: %s" % (node, e))

    def __iter__(self):
        """Iterator of node objects."""
        for node in self.nodes.values():
            yield node

    def init(self):
        """Initialize all nodes.

        Under normal circumstances, i.e. in a running guardian daemon,
        node initialization is handled automatically.  This function
        therefore does not need to be executed in user code.

        """
        for node in self:
            node.init()

    def set_managed(self, nodes=None):
        """Set all nodes to be managed by this manager.

        `names` can be a list of node names to set managed.

        """
        if nodes:
            nl = [self[node] for node in nodes]
        else:
            nl = self
        for node in nl:
            node.set_managed()

    def release(self, nodes=None):
        """Release all nodes from management by this manager.

        `nodes` can be a list of node names to release.

        """
        if nodes:
            nl = [self[node] for node in nodes]
        else:
            nl = self
        for node in nl:
            node.release()

    @property
    def arrived(self):
        """Return True if all nodes have arrived at their requested state."""
        for node in self:
            if not node.arrived:
                return False
        return True

    @property
    def completed(self):
        """Return True if all nodes are arrived and done."""
        for node in self:
            if not node.completed:
                return False
        return True

    def get_stalled_nodes(self):
        """Return a list of all stalled nodes."""
        return [node for node in self if node.STALLED]

    def revive_all(self):
        """Revive all stalled nodes."""
        for node in self:
            if node.STALLED:
                log("Reviving stalled node: %s" % node.name)
                node.revive()

    def not_ok(self):
        """Return set of node names not currently reporting OK status."""
        notok = set()
        for node in self:
            if not node.OK:
                notok.add(node.name)
        return notok

    def check_fault(self):
        """Check fault status of all nodes.

        Runs check_fault() method for all nodes.  Returns True if any
        nodes are in fault.

        """
        any_fault = False
        for node in self:
            any_fault |= node.check_fault()
        return any_fault

    @functools.wraps(node_checker)
    def checker(self, fail_return=None):
        return node_checker(self, fail_return=fail_return)



import multiprocessing
import ctypes

class SharedMemEnum(object):
    def __init__(self, enum, lock=True):
        self._enum = enum
        self._sm = multiprocessing.Value('H', lock=lock)

    @property
    def index(self):
        return self._sm.value

    @index.setter
    def index(self, index):
        self._sm.value = index

    @property
    def value(self):
        return self._enum[self._sm.value]

    @value.setter
    def value(self, value):
        try:
            self._sm.value = self._enum.index(value)
        except ValueError:
            raise ValueError("unknown SharedMemEnum index: %s" % value)


class SharedMemString(object):
    def __init__(self, size=40, lock=True):
        self._size = size
        self._sm = multiprocessing.Array(ctypes.c_char, size, lock=lock)
        self.set_value()

    @property
    def raw(self):
        return self._sm.raw

    @property
    def value(self):
        return self._sm.value.decode()

    @value.setter
    def value(self, value):
        self.set_value(value)

    def set_value(self, value=''):
        value = str(value)
        size = self._size
        self._sm.value = '{:\0<{size}}'.format(value[:size], size=size).encode()



# Versioning system that fits into an 32-bit int
# (for stuffing into an EPICS 32-bit int channel)
#
# Break up the 32-bit int into 8x4-bit (hexadecimal) fields and assign
# them thusly:
#
# 0 [reserved] (EPICS doesn't actually give the full 32 bits for the value)
# 1 [reserved]
# 2 [reserved]
# 3 MAJOR (0-15)
# 4 MINOR
# 5 MINOR (0-255)
# 6 REV
# 7 REV   (0-255)
#
# FIXME: what should we do if the version string has a suffix?


def check_vers(major, minor, rev):
    if not (major < 16**1 and minor < 16**2 and rev < 16**2):
        raise ValueError("invalid version number: out of range")


def vers_to_str(major, minor, rev):
    """return version string major.minor.rev"""
    check_vers(major, minor, rev)
    return '{0}.{1}.{2}'.format(major, minor, rev)


def vers_to_hex(major, minor, rev):
    """return version hex from major,minor,rev"""
    check_vers(major, minor, rev)
    return '000{0:1x}{1:02x}{2:02x}'.format(
        major,
        minor,
        rev,
        )


def vers_to_int(major, minor, rev, suffix=''):
    """return version int from major,minor,rev"""
    check_vers(major, minor, rev)
    return int(vers_to_hex(major, minor, rev), 16)


def vers_from_str(s):
    """return version info from version string: major,minor,rev,suffix

    major,minor,rev will be ints, where as suffix will be a str

    """
    # check if there's a suffix on the string
    suffix = ''
    vsplit = s.split('+')
    if len(vsplit) > 1:
        suffix = '+'+vsplit[1]
    try:
        major, minor, rev = vsplit[0].split('.')
    except ValueError:
        raise ValueError("Could not parse version string '{}' (should consist of major.minor.rev)".format(s))
    return int(major), int(minor), int(rev), suffix


def vers_from_int(n):
    """return version major,minor,rev from version int"""
    h = hex(n)[2:].zfill(8)
    major = int(h[3], 16)
    minor = int(h[4:6], 16)
    rev = int(h[6:8], 16)
    return major, minor, rev


def vers_str2int(s):
    """return version int from version string"""
    return vers_to_int(*vers_from_str(s))


def vers_int2str(s):
    """return version string from version int"""
    return vers_to_str(*vers_from_int(s))


def _release(s, t=None):
    """Return new release version for given version string and release type

    """
    major, minor, rev, suffix = vers_from_str(s)
    if t == 'major':
        major += 1
        minor = 0
        rev = 0
    elif t == 'minor':
        minor += 1
        rev = 0
    elif t == 'rev':
        rev += 1
    return vers_to_str(major, minor, rev)


if __name__ == '__main__':
    import argparse

    from . import __version__

    parser = argparse.ArgumentParser(
        description="""
With no arguments outputs the current version.  With a version string
or int argument, convert to int or string.  With --release option,
determine appropriate new release version from the specified version,
or from the current version if not specified.""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        'VERS', nargs='?',
        help="version string or int")
    vgroup = parser.add_mutually_exclusive_group(required=False)
    vgroup.add_argument(
        '--release', '-r', choices=['major', 'minor', 'rev'],
        help="create new release version")
    args = parser.parse_args()
    if args.release:
        vers = args.VERS or __version__
        print(_release(vers, args.release))
    elif args.VERS:
        try:
            print(vers_int2str(int(args.VERS)))
        except ValueError:
            print(vers_str2int(args.VERS))
    else:
        print(__version__)



import os
import time
import logging

def get_logger(name, level='INFO'):
    format = '%(name)s %(message)s'
    datefmt = None
    # turn off log time stamps if NOTIME is non-empty
    if os.getenv('GUARD_LOG_NOTIME', '') == '':
        format = '%(asctime)s.%(msecs)03dZ ' + format
        # time in ISO 8601 format
        # https://en.wikipedia.org/wiki/ISO_8601
        datefmt = '%Y-%m-%dT%H:%M:%S'
    # log level definitions:
    # CRITICAL: messages associated with guard termination
    # ERROR   :
    # WARNING : *** used for operator comments ***
    # INFO    : standard logging
    # DEBUG   : debugging
    # MULTI   : DEBUG + multiprocessing logging enabled
    loglevel = os.getenv('GUARD_LOG_LEVEL', level).upper()
    if loglevel == 'MULTI':
        loglevel = 'DEBUG'
        import multiprocessing
        multiprocessing.log_to_stderr().setLevel('DEBUG')
    logging.addLevelName(5, 'LOOP')
    logger = logging.getLogger(name)
    logger.setLevel(loglevel)
    formatter = logging.Formatter(format, datefmt=datefmt)
    if datefmt:
        formatter.converter = time.gmtime
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger

# FIXME: use SocketHandler to collect and serialize logs from worker subprocess
# http://docs.python.org/2/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes

class StateLogger(logging.LoggerAdapter):
    def __init__(self, logger, state=None):
        super(StateLogger, self).__init__(logger, {'state': state})
        self.extra['state'] = ''
        self.extra['func'] = ''

    def set_state(self, state):
        self.extra['state'] = state

    def set_func(self, func):
        self.extra['func'] = func

    def process(self, msg, kwargs):
        msg = '[%s.%s] %s' % (self.extra['state'], self.extra['func'], msg)
        return (msg, kwargs)



import os

from . import const

USERAPPS_DIR = os.getenv('USERAPPS_DIR', '/opt/rtcds/userapps/release').rstrip('/')

SUBSYSTEMS = [
    'als',
    'asc',
    'cal',
    'hpi',
    'ioo',
    'isc',
    'isi',
    'lsc',
    'omc',
    'psl',
    'sus',
    'sqz',
    'sys',
    'tcs',
    ]

if const.SITE and const.IFO:
    DEFAULT_ARCHIVE_ROOT = os.path.join('/ligo', 'cds',
                                        const.SITE.lower(),
                                        const.IFO.lower(),
                                        'guardian', 'archive')
else:
    DEFAULT_ARCHIVE_ROOT = None
ARCHIVE_ROOT = os.getenv('GUARD_ARCHIVE_ROOT', DEFAULT_ARCHIVE_ROOT)


class LigoPathError(Exception):
    pass

def userapps_path(*args):
    """Return full path in LIGO userapps for given specified subdirectories.

    Works like os.path.join.

    """
    return os.path.abspath(os.path.join(USERAPPS_DIR, *args))

def userapps_subpath_from_path(path):
    """For path in LIGO userapps, return subpath relative to userapps root.

    """
    path = os.path.expanduser(path).rstrip('/')
    uap = USERAPPS_DIR + '/'
    if os.path.commonprefix([path, uap]) != uap:
        raise LigoPathError('specified path is not in USERAPPS.')
    return os.path.relpath(path, USERAPPS_DIR)

def userapps_subsys_from_path(path):
    """For path in LIGO userapps, return subsystem to which it belongs.

    """
    subpath = userapps_subpath_from_path(path)
    if subpath:
        return subpath.split('/')[0]

def userapps_guardian_paths(subsyss=SUBSYSTEMS, ifo=None):
    """Return list of Guardian site-specific and common paths for list of subsystems.

    Site-specific paths will be ordered before the corresponding common path.

    If subsystem not specified, path list for all subsystems will be returned.

    If ifo is not specified, IFO environment variable will be used.

    """
    if not ifo:
        ifo = const.IFO
    if ifo:
        ifo = ifo.lower()
    paths = []
    for subsys in subsyss:
        if ifo:
            paths += [os.path.join(USERAPPS_DIR, subsys, ifo, 'guardian')]
        paths += [os.path.join(USERAPPS_DIR, subsys, 'common', 'guardian')]
    return paths



import os
import logging

from guardian.system import GuardSystem, GuardSystemError
from guardian.const import CAS_PREFIX_FMT
from guardian.db import guarddb

from .env import ENV


ISO_FMT = '%Y-%m-%d_%H:%M:%S.%f'#%z'

CHANFILE_HEADER = '''
[default]
gain=1.00
datatype=4
ifoid=0
slope=1.00
acquire=3
offset=0
units=V
dcuid=4
datarate=16
'''


class GuardCtrlError(Exception):
    pass


def check_system(name):
    try:
        GuardSystem(name)
    except GuardSystemError as e:
        raise GuardCtrlError(e)


def check_vars():
    if not ENV['GUARD_CHANFILE']:
        return False
    if not ENV['IFO']:
        raise GuardCtrlError("IFO variable must be specified for writing guard channel file.")
    return True


def test_chanfile():
    if not check_vars():
        return
    CHANFILE = ENV['GUARD_CHANFILE']
    try:
        open(CHANFILE, 'a').close()
    except Exception as e:
        raise GuardCtrlError(e)


def update_global_chanfile(node_list):
    if not check_vars():
        return
    IFO = ENV['IFO']
    CHANFILE = ENV['GUARD_CHANFILE']
    logging.info("updating global channel file: {}".format(CHANFILE))
    cdir = os.path.dirname(CHANFILE)
    if not os.path.exists(cdir):
        os.makedirs(cdir)
    tfile = CHANFILE+'.new'
    with open(tfile, 'w') as f:
        # f.write(CHANFILE_HEADER.strip())
        f.write('\n\n')
        for node in node_list:
            for chan, attr in sorted(guarddb.items()):
                if not attr.get('archive'):
                    continue
                channel = CAS_PREFIX_FMT.format(IFO=IFO, SYSTEM=node) + chan
                f.write('[{}]\n'.format(channel))
    os.rename(tfile, CHANFILE)



import os
import logging
import numpy as np
import matplotlib.pyplot as plt
# from matplotlib.scale import ScaleBase, register_scale
from matplotlib import axes
from matplotlib.axes import Axes
from matplotlib.gridspec import GridSpec
from matplotlib.ticker import MultipleLocator, FixedLocator, NullLocator
from matplotlib.ticker import FuncFormatter
from matplotlib.colors import ListedColormap
#from matplotlib.patches import Rectangle

from gpstime import gpstime

from pydv.util import make_default_logger
from pydv import NDSBufferDict, NDSBuffer

from . import const
from .db import guarddb

##################################################

DEFAULT_LOG_LEVEL = 'DEBUG'

NODE_CHANNEL_SUFFIX = [
    'STATE_N',
    'REQUEST_N',
    'NOMINAL_N',
    'OP',
    'MODE',
    'ERROR',
    'CONNECT',
    'LOAD_STATUS',
    'NOTIFICATION',
    'STATUS',
    'WORKER',
    'EXECTIME',
    'OK',
    ]

##################################################

class GuardAxes(Axes):
    logger = make_default_logger('GuardAxes')

    def __init__(self, fig, rect,
                 center_time=None,
                 xmin=None,
                 xmax=None,
                 channel_names=None,
                 ndsbufferdict=None,
                 transform=lambda x: x,
                 enum_colors=False,
                 initial_plot_kwargs={},
                 figure_updated_flag_sem_pair=None,
                 **kwargs):
        Axes.__init__(self, fig, rect, **kwargs)
        self.center_time = center_time
        self.ndsbufferdict = ndsbufferdict
        self.transform = transform
        self.enum_colors = enum_colors
        self.handles = {}
        self.__last_xlim = self.get_xlim()
        self.__last_ylim = self.get_ylim()
        self.__data_ylim = []

        #self.ndsbufferdict.add_channels(channel_names)
        for channel in channel_names:
            self.update_data(channel, **initial_plot_kwargs)

        self.__last_ylim = self.get_ylim()

        self.grid(True)

        self.figure.canvas.mpl_connect('button_release_event', self.get_more_data)

    #def get_ax_size(self):
    def get_ax_height(self):
        return 1000
        # fig = self.figure
        # #bbox = self.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
        # trans = self.transData.inverted()
        # h = trans.transform((0,0)) - trans.transform((0,1))
        # #width, height = bbox.width, bbox.height
        # width, height = size
        # width *= fig.dpi
        # height *= fig.dpi
        # return width, height

    def update_data(self, channel, **kwargs):
        ndsbuffer = self.ndsbufferdict[channel]
        channel_name = ndsbuffer.channel_name
        time_domain = ndsbuffer.time_domain - self.center_time
        data = self.transform(ndsbuffer.data)
        if self.enum_colors:
            ydata = [0]*len(data)
            colors = self.enum_colors
            cmap = ListedColormap(colors)
            #cmap = plt.get_cmap('jet', 3)
            # FIXME: look at:
            # p = patches.Rectange
            # ax.add_patch(p)
            handle = self.scatter(time_domain, ydata, c=data,
                                  cmap=cmap, vmin=0, vmax=len(self.enum_colors)-1,
                                  label=channel,
                                  s=self.get_ax_height(),
                                  marker=((1,4), (1,-4), (-1,-4), (-1,4)),
                                  linewidth=0,
                                  **kwargs)
            if channel_name in self.handles:
                self.handles[channel].remove()
            self.handles[channel_name] = handle
        else:
            if channel_name not in self.handles:
                handle = self.plot(time_domain, data,
                                   label=channel,
                                   **kwargs)[0]
                #picker=self.line_picker, **kwargs)[0]
                self.handles[channel_name] = handle
            else:
                handle = self.handles[channel_name]
                handle.set_xdata(time_domain)
                handle.set_ydata(data)
        self.update_limits(data)

    def update_limits(self, data):
        if self.__data_ylim:
            ymin, ymax = self.__data_ylim
            ymin = min(ymin, min(data))
            ymax = max(ymax, max(data))
        else:
            ymin = min(data)
            ymax = max(data)
        self.__data_ylim = [ymin, ymax]

    @property
    def channels(self):
        return self.handles.keys()

    def get_more_data(self, event):
        xlim = self.get_xlim()
        if self.__last_xlim != xlim:
            self.logger.debug("x limits have changed. Adding more data...")
            self.__last_xlim = xlim

            xmin, xmax = xlim
            xmin += self.center_time
            xmax += self.center_time
            self.ndsbufferdict.add_data(xmin, xmax)

            # for channel in self.handles:
            #     self.update_data(channel)

        if not self.enum_colors:
            if self.__last_ylim == self.get_ylim():
                self.set_ylim(self.__data_ylim)
            self.__last_ylim = self.get_ylim()
        #self.relim()
        self.figure.canvas.draw()

GuardSubplot = axes.subplot_class_factory(GuardAxes)

class GuardDataVfunc(object):
    def __init__(self, ind, unki):
        self.ind = ind
        self.unki = unki
    def __call__(self, index):
        return self.ind.get(index, self.unki)

class GuardPlotTransform(object):
    unknown_ind = -2

    def __init__(self, system):
        self.system = system

        self.index_n_dict = {-1: -1,
                             -2: -2,
                             }
        self.n_label_dict = {-2: '????',
                             -1: 'NONE (-1)',
                             }
        #isd = system.index_state_dict()
        #for n, d in enumerate([(isd[i], i) for i in sorted(isd)]):
        for n, d in enumerate(list(system.states_iter(reverse=False))):
            self.index_n_dict[d[1]] = n
            self.n_label_dict[n] = '%s (%d)' % (d[0], d[1])

        #self.index_n_vfunc = np.vectorize(self.index_n_dict.__getitem__)
        self.index_n_vfunc = np.vectorize(GuardDataVfunc(self.index_n_dict, self.unknown_ind))

    def transform_data(self, data):
        return self.index_n_vfunc(data)

    def get_label(self, n, pos=None):
        return '   %s' % self.n_label_dict.get(n, 'FOO')

    def transform_axis(self, ax):
        #ax.yaxis.set_major_locator(MultipleLocator(1))
        ax.yaxis.set_major_locator(FixedLocator(self.index_n_dict.values()))
        ax.yaxis.set_major_formatter(FuncFormatter(self.get_label))

        for tick in ax.yaxis.get_ticklabels():
            tick.set_horizontalalignment('left')
            tick.set_verticalalignment('center')
            #tick.set_position((0.02, 0)) #tick.get_position()[1]))

##################################################

def add_colorbar_axis(fig, gs, prefix, channel, colors=[], line_width=50, **kwargs):
    ax = GuardSubplot(fig, gs,
                      channel_names = ['{0}{1}'.format(prefix, channel)],
                      enum_colors=colors,
                      **kwargs)
    ax.set_title('', visible=False)
    #ax.legend().remove()
    ax.grid(False)
    ax.yaxis.set_major_locator(NullLocator())
    # HACK: this is how you hide tick labels on one axis when sharex
    # is in effect
    # https://stackoverflow.com/questions/4209467/matplotlib-share-x-axis-but-dont-show-x-axis-tick-labels-for-both-just-one
    #plt.setp(ax.get_xticklabels(), visible=False)
    ax.label_outer()
    ax.set_ylabel(channel, rotation=0, va='center', ha='right')
    fig.add_subplot(ax)
    return ax

def add_enum_axis(fig, gs, prefix, channel, line_width=2, **kwargs):
    def formatter(n, *args):
        sd = {n: s for n,s in enumerate(guarddb[channel]['enums'])}
        label = sd.get(n, '')
        if label:
            return '   {0}'.format(label)
        else:
            return label
    ax = GuardSubplot(fig, gs,
                      channel_names = ['{0}{1}'.format(prefix, channel)],
                      **kwargs)
    ax.set_title('', visible=False)
    #ax.legend().remove()
    #plt.setp(ax.get_xticklabels(), visible=False)
    ax.label_outer()
    ax.yaxis.set_major_locator(MultipleLocator(1))
    ax.yaxis.set_major_formatter(FuncFormatter(formatter))
    ax.set_ylim([-1, len(guarddb[channel]['enums'])])
    ax.set_ylabel(channel, rotation=0, va='center', ha='right')
    for tick in ax.yaxis.get_ticklabels():
        tick.set_horizontalalignment('left')
        tick.set_verticalalignment('center')
        #tick.set_position((0.02, 0))
    # for line in ax.lines:
    #     line.set_linewidth(line_width)
    fig.add_subplot(ax)
    return ax

def add_state_axis(fig, gs, prefix, system, **kwargs):
    transform = GuardPlotTransform(system)
    ax = GuardSubplot(fig, gs,
                       channel_names=['{0}{1}_N'.format(prefix, chan) for chan in ['STATE', 'REQUEST', 'NOMINAL']],
                       transform=transform.transform_data,
                       **kwargs)
    ax.set_title('', visible=False)
    transform.transform_axis(ax)
    ax.set_xlabel('seconds after %f (%s)' % (kwargs['center_time'],
                                             gpstime.fromgps(kwargs['center_time']).isoformat(' ')),
                  ha='center')
    ax.set_ylabel('STATE', rotation=0, va='center', ha='right')
    ax.lines[0].set_color('gray')
    ax.lines[0].set_linewidth(12)
    ax.lines[0].set_alpha(0.5)
    ax.lines[1].set_color('yellow')
    ax.lines[1].set_linewidth(10)
    ax.lines[1].set_alpha(0.5)
    ax.lines[2].set_color('green')
    ax.lines[2].set_linewidth(8)
    ax.lines[2].set_alpha(0.5)
    ax.legend(['STATE', 'REQUEST', 'NOMINAL'], loc='lower right')
    fig.add_subplot(ax)
    return ax

##################################################

def guardfig(system, center_time, window):

    prefix = const.CAS_PREFIX_FMT.format(IFO=system.ifo, SYSTEM=system.name)

    def _addHandler(logger):
        formatter = logging.Formatter('%(asctime)s : %(name)s : %(message)s')
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        handler.set_name('tests')
        logger.addHandler(handler)
        logger.setLevel(os.getenv('LOG_LEVEL', 'DEBUG').upper())

    _addHandler(NDSBufferDict.logger)
    _addHandler(NDSBuffer.logger)

    center_time = gpstime.parse(center_time).gps()
    xmin = center_time + window[0]
    xmax = center_time + window[1]

    ndsbufferdict = NDSBufferDict(['{0}{1}'.format(prefix, chan) for chan in NODE_CHANNEL_SUFFIX], xmin, xmax)

    kwargs = dict(center_time=center_time,
                  xmin=xmin,
                  xmax=xmax,
                  ndsbufferdict=ndsbufferdict,
                  )

    fig = plt.figure()
    gs = GridSpec(50, 1)
    gs.update(right=0.98, bottom=0.05, top=.95, hspace=0)

    ax0 = add_state_axis(fig, gs[23:,0], prefix, system, **kwargs)
    kwargs['sharex'] = ax0
    add_enum_axis(fig, gs[7:13,0], prefix, 'STATUS', **kwargs)
    add_enum_axis(fig, gs[13:19,0], prefix, 'WORKER', line_width=2, **kwargs)
    add_colorbar_axis(fig, gs[0,0], prefix, 'OP', colors=['#FF6600','#FFFF00','#00CC00'], **kwargs)
    add_colorbar_axis(fig, gs[1,0], prefix, 'MODE', colors=['#0000FF','#9900CC','#FF99FF'], **kwargs)
    add_colorbar_axis(fig, gs[2,0], prefix, 'ERROR', colors=['#00CC00','#FF0000'], **kwargs)
    add_colorbar_axis(fig, gs[3,0], prefix, 'CONNECT', colors=['#00CC00','#FF0000'], **kwargs)
    add_colorbar_axis(fig, gs[4,0], prefix, 'LOAD_STATUS', colors=['#00CC00','#FF0000'], **kwargs)
    add_colorbar_axis(fig, gs[5,0], prefix, 'NOTIFICATION', colors=['#00CC00','#FFFF00'], **kwargs)
    add_colorbar_axis(fig, gs[6,0], prefix, 'OK', colors=['#FF6600','#00CC00'], **kwargs)
    #add_enum_axis(fig, gs[10:14,0], prefix, 'WORKER', line_width=2, **kwargs)
    #add_enum_axis(fig, gs[10:14,0], prefix, 'OP', **kwargs)
    # fig.add_subplot(GuardSubplot(fig, gs[10:14, 0],
    #                              channel_names = ['{0}{1}'.format(prefix, 'MODE')],
    #                              **kwargs))

    ax_et = GuardSubplot(fig, gs[19:23, 0],
                         channel_names = ['{0}{1}'.format(prefix, 'EXECTIME')],
                         **kwargs)
    ax_et.set_title('', visible=False)
    #ax_et.legend().remove()
    ax_et.label_outer()
    ax_et.set_ylabel('EXECTIME', rotation=0, va='center', ha='right')
    #ax_et.autoscale(tight=True)
    fig.add_subplot(ax_et)


    #ax0.axis('tight')
    ax0.set_xlim(window)

    #gs.tight_layout(fig)

    fig.suptitle("{0} GRD: {1}".format(system.ifo, system.name))

    plt.show()

    return fig



from . import version
from ._version import __version__
from . import const

DAEMON_OP = ('STOP', 'PAUSE', 'EXEC')
DAEMON_MODE = ('AUTO', 'MANAGED', 'MANUAL')
DAEMON_STATUS = ('ENTER', 'MAIN', 'RUN', 'DONE', 'JUMP', 'REDIRECT', 'EDGE', 'INIT')
WORKER_STATUS = ('INIT', 'COMMAND', 'CONTINUE', 'DONE', 'JUMP', 'CERROR', 'ERROR')

############################################################
############################################################
# Guardian channel database definition
#
# This lists all channels that are used internally by the Guardian
# Daemon.
#
# It is also the database definition for the pcaspy EPICS portable
# channel access server:
#
#   https://pcaspy.readthedocs.org/en/latest/api.html#id1
#
# It lists all channels and their associated data types.

guarddb = {
    # request state
    # enum set at load time with states from system.
    'REQUEST': {
        'type': 'string',
        'guardstate': True,
        'writable': True,
    },
    # current state
    'STATE': {
        'type': 'string',
        'guardstate': True,
    },
    # target state
    'TARGET': {
        'type': 'string',
        'guardstate': True,
    },
    # nominal state
    'NOMINAL': {
        'type': 'string',
        'guardstate': True,
    },

    # op mode
    'OP': {
        'type': 'enum',
        'enums': DAEMON_OP,
        'writable': True,
        'archive': True,
    },
    # run mode
    'MODE': {
        'type': 'enum',
        'enums': DAEMON_MODE,
        'writable': True,
        'archive': True,
    },

    # load request
    'LOAD': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'writable': True,
    },
    'LOAD_STATUS': {
        'type': 'enum',
        'enums': ('DONE', 'REQUEST', 'SET', 'INPROGRESS', 'ERROR'),
        'archive': True,
    },

    # management
    'MANAGER': {
        'type': 'string',
        'value': '',
        'writable': True,
    },
    'STALLED': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },

    # code execution status
    'STATUS': {
        'type': 'enum',
        'enums': DAEMON_STATUS,
        'archive': True,
    },
    # worker return status
    'WORKER': {
        'type': 'enum',
        'enums': WORKER_STATUS,
    },
    # worker execution time and cycles
    'EXECTIME': {
        'type': 'float',
        'unit': 'seconds',
        'value': -1,
        'prec': 4,
        'lolo': -1,
        'low': -1,
        'high': 1./const.CPS,
        'hihi': 1,
        'log_changes': False,
        'archive': True,
    },
    'EXECCYCLES': {
        'type': 'int',
        'value': -1,
        'lolo': -1,
        'low': -1,
        'high': 1,
        'hihi': const.CPS,
        'log_changes': False,
    },
    'EXECTIME_LAST': {
        'type': 'float',
        'unit': 'seconds',
        'value': -1,
        'prec': 4,
        'lolo': -1,
        'low': -1,
        'high': 1./const.CPS,
        'hihi': 1,
        'log_changes': False,
    },
    'EXECCYCLES_LAST': {
        'type': 'int',
        'value': -1,
        'lolo': -1,
        'low': -1,
        'high': 1,
        'hihi': const.CPS,
        'log_changes': False,
    },
    'LOOPUSEC': {
        'type': 'int',
        'log_changes': False,
    },
    'MAINUSEC': {
        'type': 'int',
        'log_changes': False,
    },

    # error flag
    'ERROR': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },
    # connection status
    'CONNECT': {
        'type': 'enum',
        'enums': ('OK', 'ERROR'),
        'archive': True,
    },

    'PV_TOTAL': {
        'type': 'int',
        'value': -1,
        'archive': True,
    },

    # manager subordinate node count
    'SUBNODES_TOTAL': {
        'type': 'int',
        'value': -1,
    },
    'SUBNODES_NOT_OK': {
        'type': 'int',
        'value': -1,
        'low': -1,
        'high': 1,
    },

    # daemon message
    'GRDMSG': {
        'type': 'char',
        'count': 100,
        'value': [0],
    },
    # user message
    'USERMSG': {
        'type': 'char',
        'count': 100,
        'value': [0],
    },
    # notification flag (USERMSG present)
    'NOTIFICATION': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },
    # operational summary bit (OP/MODE/ERROR status nominal)
    'ACTIVE': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },
    # subnode summary bit (all sub nodes OK)
    'READY': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },
    # intent summary bit (REQUEST == NOMINAL)
    'INTENT': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },
    # overall summary bit (READY + INTENT + NOMINAL)
    'OK': {
        'type': 'enum',
        'enums': ('False', 'True'),
        'guardbool': True,
        'archive': True,
    },
    # log level
    'LOGLEVEL': {
        'type': 'enum',
        'enums': ('INFO', 'DEBUG', 'LOOP'),
        'writable': True,
    },
    # daemon initialization time, in UNIX seconds
    'TIME_INIT': {
        'type': 'int',
    },
    # uptime
    'TIME_UP': {
        'type': 'int',
        'unit': 'seconds',
        'log_changes': False,
    },
    # code archive id
    'ARCHIVE_ID': {
        'type': 'int',
        'archive': True,
    },
    # system hostname
    'HOSTNAME': {
        'type': 'str',
        'value': const.HOSTNAME,
    },
    # guardian version
    'VERSION': {
        'type': 'int',
        'value': version.vers_str2int(__version__),
        'archive': True,
    },
    'VERSION_S': {
        'type': 'str',
        'value': __version__,
    },
}

# add the guardstate readback channels
# each channel marked 'guardstate' will have corresponding _S string
# and _N numeric channels added.  These will be updated according to
# a states:index dictionary
for channel, data in [tup for tup in guarddb.items()]:
    if data.get('guardstate', False):
        guarddb[channel+'_S'] = {
            'type': 'string',
        }
        guarddb[channel+'_N'] = {
            'type': 'int',
            'archive': True,
        }

# first 10 user messages
for i in range(const.USERMSG_COUNT):
    guarddb['USERMSG'+str(i)] = {
        'type': 'char',
        'count': const.USERMSG_STRING_LENGTH,
        'value': [0],
    }

# add setpoint monitor channels
# whether monitoring is enabled
guarddb['SPM_MONITOR'] = {
    'type': 'enum',
    'enums': ('False', 'True'),
    'guardbool': True,
    'archive': True,
}
guarddb['SPM_MONITOR_NOTIFY'] = {
    'type': 'enum',
    'enums': ('False', 'True'),
    'guardbool': True,
}
# total number of setpoints
guarddb['SPM_TOTAL'] = {
    'type': 'int',
    'value': -1,
    'archive': True,
}
# number of setpoints that have changed
guarddb['SPM_CHANGED'] = {
    'type': 'int',
    'value': -1,
    'lolo': -1,
    'low': -1,
    'high': 1,
    'hihi': 1,
    'archive': True,
}
# first 10 setpoint differences
for i in range(const.SPM_DIFF_COUNT):
    # channel name
    guarddb['SPM_DIFF'+str(i)] = {
        'type': 'char',
        'count': const.CHANNEL_NAME_LENGTH,
        'value': [0],
    }
    for t in ['S', 'C', 'D']:
        # setpoint value, string representation
        guarddb['SPM_DIFF'+str(i)+'_'+t] = {
            'type': 'string',
            'value': '',
        }
# SPM snapshot
guarddb['SPM_SNAP'] = {
    'type': 'enum',
    'enums': ('False', 'True'),
    'guardbool': True,
    'writable': True,
}

############################################################
############################################################

class Database(object):
    def __init__(self, cas):
        self._cas = cas
        self._db = {}
        self._writable = []
        for channel, entry in guarddb.items():
            self._db[channel] = None
            if entry.get('writable', False):
                self._writable.append(channel)
            if 'value' in entry:
                self._db[channel] = entry['value']
            elif entry['type'] == 'enum' and entry['enums']:
                # zeroth value of enum is always default
                value = entry['enums'][0]
                if entry.get('guardbool', False):
                    value = eval(value)
                self._db[channel] = value
        self.request_event = False

    ########################################

    def __getitem__(self, channel):
        return self._db[channel]

    def __setitem__(self, channel, value):
        self._db[channel] = value
        if self._cas:
            self._cas[channel] = value

    ########################################
    # CAS interaction

    def update(self):
        """Update the db from the cas.

        """
        if not self._cas:
            return

        if self._cas._request_event.is_set():
            self.request_event = True
            self._cas._request_event.clear()

        for channel in self._writable:
            self._db[channel] = self._cas[channel]


    def update_system(self, system, init=False):
        """Update system definition

        E.g. state index and request state enum

        """
        if not self._cas:
            return
        index = system.state_index_dict()
        # FIXME: this truncation should throw a WARNING
        enum = system.request_states[:const.REQUEST_STATE_LIMIT]
        # always add NONE state
        index['NONE'] = dict(const.RESERVED_STATES)['NONE']
        # enum += ('NONE',)
        self._cas.update_state_index(index, init=init)
        return self._cas.update_request_enum(enum, init=init)



import os
import socket

# SITE and IFO designators
SITE = os.getenv('SITE', '')
IFO = os.getenv('IFO', '')

# system hostname
HOSTNAME = socket.gethostname()

# daemon EPICS channel access server channel prefix
CAS_PREFIX_FMT = '{IFO}:GRD-{SYSTEM}_'

# main daemon step time in Hz
CPS = int(os.getenv('GUARD_CPS', 16))
if 1000000 % CPS != 0:
    raise ValueError("CPS does not correspond to an integer number of microseconds")

# how long to wait, in seconds, on REDIRECT before terminating worker
REDIRECT_TIMEOUT = 1

# channel names length
CHANNEL_NAME_LENGTH = 60

# USERMSG char array length
USERMSG_STRING_LENGTH = 100

# number of USERMSG channels
USERMSG_COUNT = 10

# how frequently to step through USERMSG
USERMSG_CYCLE_TIME = 2

# number of SPM_DIFF channels
SPM_DIFF_COUNT = 10

# maximum acceptable length of state names.
# (39 characters is the EPICS string length limit)
STATE_NAME_LENGTH = 39

# list of reserved state names and indices
RESERVED_STATES = [
    ('NONE', -1),
    ('????', -2),
    ('????', -3),
    ('????', -4),
    ('????', -5),
    ('????', -6),
    ('????', -7),
    ('????', -8),
    ('????', -9),
]

# default weight of goto edges
DEFAULT_GOTO_EDGE_WEIGHT = 1

# number of allowable REQUEST states
# HACK: EPICS allows for only 16 element enums, and the NONE state is
# always added, so 15 is the limit
REQUEST_STATE_LIMIT = 15



import os
import re
import gzip
import fnmatch
import logging
import subprocess
from gpstime import gpstime
from dateutil.tz import tzutc, tzlocal

from . import util

##################################################

GUARD_TARGET = os.getenv('GUARD_TARGET', '/etc/guardian')
GUARD_NODE_DIR = os.path.join(GUARD_TARGET, 'nodes')
GUARD_NODE_LOG_DIR = os.path.join(GUARD_TARGET, 'logs')

TIME_RE = re.compile('(\d{4}-?\d{2}-?\d{2}[T_]\d{2}:\d{2}:\d{2}\S*) ?(.*)')

##################################################

def full_node_list():
    """return list of all active nodes"""
    return os.listdir(GUARD_NODE_DIR)


def nodes_expand(glob_list=None):
    """return list of nodes matching patterns

    Match a list of globbed node names with the full node list,
    returning a list of matching nodes.  If no glob list is provided
    all nodes will be returned.

    An exception is raised if a pattern matches no known nodes.

    """
    node_list = full_node_list()
    if not glob_list:
        return sorted(node_list)
    expanded = set()
    for glob in glob_list:
        matches = fnmatch.filter(node_list, glob)
        if not matches:
            raise util.GuardCtrlError("Pattern {} matches no known nodes".format(glob))
        expanded |= set(matches)
    return sorted(expanded)

##########

def in_interval(t, after=None, before=None):
    """determine if time within interval"""
    ab = True
    bb = True
    if after and t < after:
        ab = False
    if before and t > before:
        bb = False
    return ab and bb

##########

def opportunistically_decompress(filename, mode='r'):
    """trick to opportunistically decompress gzip files

    Opens filename and looks for the gzip "magic number" in the first
    couple of bytes.  If it is not found the open file object is
    returned.  If it is found, the file object is un-gziped and then
    returned.

    """
    GZIP_MAGIC_NUMBER = '\x1f\x8b\x08'
    fd = open(filename, mode)
    # gzip magic number
    data = fd.read(len(GZIP_MAGIC_NUMBER))
    fd.seek(0)
    if data != GZIP_MAGIC_NUMBER:
        return fd
    else:
        return gzip.GzipFile(filename, mode, fileobj=fd)

##################################################

def find_log_files(logdir):
    """generator of all svlogd log files in directory"""
    # for python3:
    # with subprocess.Popen(['/usr/bin/tai64nlocal'],
    #                       stdin=subprocess.PIPE,
    #                       stdout=subprocess.PIPE,
    #                       #universal_newlines=True,
    #                       ) as p:
    # FIXME: this is obviously non-portable
    p = subprocess.Popen(['/usr/bin/tai64nlocal'],
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         )
    if True:
        for f in os.listdir(logdir):
            if f[0] != '@':
                continue
            i = f.split('.')[0]
            p.stdin.write(i+'\n')
            o = p.stdout.readline().strip().split('.')[0]
            dt = gpstime.strptime(o, '%Y-%m-%d %H:%M:%S')
            # tai64n log file names are local time
            dt = dt.replace(tzinfo=tzlocal())
            yield (f, dt)
        p.terminate()
        p.wait()
        if os.path.exists(os.path.join(logdir, 'current')):
            # include 'current' at the end
            yield ('current', gpstime.now(tz=tzlocal()))

def filter_files(logdir, after=None, before=None):
    """generator of svlogd log files from directory covering interval (after,before)"""
    last = gpstime.fromtimestamp(0, tz=tzutc())
    for f, dt in sorted(find_log_files(logdir), key=lambda k: k[1]):
        bb = True
        ab = True
        if after and after > dt:
            bb = False
        if before and before < last:
            ab = False
        last = dt
        if bb and ab:
            yield (f, dt)

def filter_files_multidir(logdirs, after=None, before=None):
    """generator of all svlogd files in a list of directories, filtered by time interval"""
    for logdir in logdirs:
        for f, dt in filter_files(logdir, after, before):
            yield (os.path.join(logdir, f), dt)

def parse_log_line(line):
    """parse log line

    Take a log line string and return a tuple of (datetime, msg)
    extracted from the line.

    """
    line = line.strip()
    match = TIME_RE.match(line)
    if not match:
        return None, line
    ds = match.group(1)
    msg = match.group(2)
    # dt = gpstime.strptime(ds.split('.')[0], '%Y-%m-%dT%H:%M:%S')
    try:
        dt = gpstime.strptime(ds, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        try:
            dt = gpstime.strptime(ds, '%Y-%m-%dT%H:%M:%S.%fZ')
        except ValueError:
            try:
                dt = gpstime.strptime(ds, '%Y%m%d_%H:%M:%S.%f')
            except:
                return None, msg
    # guardian log times are UTC
    dt = dt.replace(tzinfo=tzutc())
    return dt, msg

def dump_lines_multifile(fs):
    """generator of log lines from specified files"""
    for f in fs:
        logging.debug("dumping file: %s" % f)
        with opportunistically_decompress(f) as fd:
            for line in fd:
                yield line.strip()

def filter_lines(lines, after=None, before=None):
    """filter lines within specified range"""
    for line in lines:
        dt, msg = parse_log_line(line)
        if dt:
            if in_interval(dt, after, before):
                yield (dt, msg)
        else:
            yield (dt, msg)

def filter_lines_multifile(fs, after=None, before=None):
    """generator of log lines within specified range from specified files"""
    for f in fs:
        logging.debug("filtering file: %s" % f)
        with opportunistically_decompress(f) as fd:
            for log in filter_lines(fd, after, before):
                yield log

##################################################

def gen_node_log_dirs(nodes):
    """generator of node log directory paths"""
    for node in nodes:
        yield os.path.join(GUARD_NODE_LOG_DIR, node)

def find_node_log_files(nodes, after=None, before=None):
    """find all log files for the specified nodes within time range"""
    nodedirs = list(gen_node_log_dirs(nodes))
    if after is True:
        return [os.path.join(d, 'current') for d in nodedirs]
    else:
        return [p for p,dt in sorted(filter_files_multidir(nodedirs, after, before), key=lambda k: k[1])]

def sort_log_lines(lines, sort='time'):
    """sort log lines by date"""
    if sort == 'time':
        return sorted(lines, key=lambda k: k[0])
    else:
        return lines

def format_log_line(dt, msg, time_format='utc'):
    """format output of log (datetime, message) tuples"""
    if dt:
        if time_format == 'utc':
            dts = dt.strftime('%Y-%m-%d_%H:%M:%S.%fZ')
        elif time_format == 'local':
            dts = dt.astimezone(tzlocal()).strftime('%Y-%m-%d_%H:%M:%S.%f')
        elif time_format == 'gps':
            dts = '{:.3f}'.format(dt.gps())
        else:
            dts = str(dt)
    else:
        dts = '??????????'
    return '{} {}'.format(dts, msg)

##################################################

def node_log_search(nodes, after=None, before=None):
    """generator of node log lines in some time range

    """
    paths = find_node_log_files(nodes, after, before)

    # opportunistically sort the lines only if more then one node is
    # specified
    if len(nodes) == 1:
        sort = None
    else:
        sort = 'time'

    it = filter_lines_multifile(paths, after, before)
    return sort_log_lines(it, sort)



# -*- fill-column: 80; -*-

import os
import sys
import imp
import inspect
import networkx
from collections import defaultdict

import builtins
import importlib
from importlib import reload
_builtin__import__ = importlib.__import__

from . import const
from .ligopath import userapps_guardian_paths
from .state import GuardState
from .manager import Node, NodeManager

##################################################

def _check_modpath(path):
    path = path.rstrip('c')
    if os.path.exists(path+'c') and not os.path.exists(path):
        msg = "Module .pyc found without corresponding .py:\n\n"
        msg += '  %sc' % path
        msg += "\n\nRemove defunct module .pyc to continue."
        raise GuardSystemImportError(msg)

##################################################

class GuardSystemError(Exception):
    pass

class GuardSystemImportError(GuardSystemError):
    pass

class GuardSystemLoadError(GuardSystemError):
    pass

class GuardSystem(object):
    """Guardian System description class.

    If not provied, IFO and GUARD_MODULE_PATH will be set from environment.

    """
    def __init__(self,
                 module=None,
                 ifo=None,
                 name=None,
                 ca_prefix=None,
                 guardpath=None,
                 ):
        if not ifo:
            ifo = const.IFO
        self._ifo = ifo
        self._name = name
        self._ca_prefix = ca_prefix

        # resolve the module search path
        if guardpath:
            assert type(guardpath) is list, "guardpath must be a list"
        else:
            guardpath = os.getenv('GUARD_MODULE_PATH')
            if guardpath:
                guardpath = guardpath.split(':')
            else:
                guardpath = userapps_guardian_paths(ifo=self._ifo)

        self._modname = None
        self._modpath = None
        self._usercode = []
        self._module = None

        if module:
            modname = module

            # if the module is specified by path, add its directory to
            # the front of the search path only if it's not already
            # included (this makes sure we don't reprioritize the
            # search path when pointing directly to modules already in
            # the search path.  Also set the module name to be the
            # basename.
            if os.path.exists(module):
                modname = os.path.splitext(os.path.basename(module))[0]
                modpath = os.path.abspath(os.path.dirname(module))
                if modpath not in guardpath:
                    guardpath.insert(0, modpath)

            # find the module, without loading it
            try:
                fp, modpath, description = imp.find_module(modname, guardpath)
                fp.close()
            except ImportError:
                ppath = ''
                for path in guardpath:
                    ppath += '\n' + path
                msg = "Module '%s' not found in any of the following search paths:%s" % (module, ppath)
                raise GuardSystemImportError(msg)

            _check_modpath(modpath)

            self._modname = modname
            self._modpath = os.path.abspath(modpath)

        self._guardpath = guardpath

        if not self._name:
            self._name = self._modname

        # these system parameters will not be reset on reload:
        self._ca_prefitx = ca_prefix

        self._reset()

    ##########

    def _reset(self):
        # resets system to it's pre-load state.

        # (re)initialize the graph object
        self._graph = networkx.DiGraph()

        # reset parameters loaded from module
        self._initial_request = None
        self._nominal_state = None
        self._manager = None
        self._node_managers = {}
        # don't recorgnize ca_prefix changes on reset
        self._ca_monitor = False
        self._ca_monitor_notify = True
        self._ca_channels = []
        self._related_displays = []

        # purge the module object
        if self._module:
            from sys import modules
            del self._module
            self._module = None
            del modules[self._modname]
            # for mod in modules.values():
            #     try:
            #         delattr(mod, self._modname)
            #     except AttributeError:
            #         pass

    def _import(self, name, *args, **kwargs):
        # custom import function that uses the builtin import function to import
        # modules, appending "usercode" modules in usercode attribute.  it also
        # reloads modules in the guardpath if they've already been loaded
        # before.  It then returns the module object itself.
        #
        # FIXME: this is all very hacky because the import internals are
        # inconsistent.  This should all be much easier in python3

        # FIXME: we need to use the builtin import because it appears
        # to be the only thing that handles both submodule import and
        # the "from . import ." form
        module = _builtin__import__(name, *args, **kwargs)

        # ignore "built-in" modules (i.e. without __file__)
        if not hasattr(module, '__file__'):
            return module

        modpath = module.__file__.rstrip('c')

        # only care about modules/packages in the guardpath
        if not self._in_guardpath(modpath):
            return module

        modpaths = set()

        # always add __file__, which will be __init__ in the case of a package
        modpaths.add(modpath)

        # generator for paths of all packages of a module
        def packageup(d):
            d = os.path.split(d)[0]
            p = os.path.join(d, '__init__.py')
            if not os.path.exists(p):
                return
            for dd in packageup(d):
                yield dd
            yield p

        # add all packages to which this module belongs
        for d in packageup(modpath):
            modpaths.add(d)

        # if module has __path__ list then it's a package from which something
        # is being imported
        if hasattr(module, '__path__'):
            for pkgpath in module.__path__:
                # HACK: direct sub-sub-module imports ('import foo.bar.baz') are
                # not detected in any other way in this framework, so we have to
                # specially look for that case and add them manually
                ssm = os.path.join(pkgpath, *name.split('.')[1:])+'.py'
                if os.path.exists(ssm):
                    modpaths.add(ssm)

                # if there's not a fromlist, then just look for modules and
                # packages in the package namespace
                if 'fromlist' in kwargs:
                    fromlist = kwargs['fromlist']
                else:
                    fromlist = [mod for mod in dir(module) if mod[:2] != '__']

                for mod in fromlist:
                    # add submodules
                    modpath = os.path.join(pkgpath, mod+'.py')
                    if os.path.exists(modpath):
                        modpaths.add(modpath)
                        continue
                    # add subpackages
                    modpath = os.path.join(pkgpath, mod, '__init__.py')
                    if os.path.exists(modpath):
                        modpaths.add(modpath)

        # now record and/or reload all found modules
        for path in modpaths:
            # resolve full module file path
            abspath = os.path.abspath(path)
            # check that this isn't a remnant .pyc
            _check_modpath(abspath)
            # add it the to usercode list
            if abspath not in [self._modpath] + self._usercode:
                self._usercode.append(abspath)
            # reload the module if this is a re-import
            if not self.__already_loaded[abspath]:
                reload(module)
                self.__already_loaded[abspath] = True

        return module

    def _in_guardpath(self, modpath):
        # return True if specified path is a subdirectory of a path in the
        # guardpath.
        modpath = os.path.abspath(modpath)
        for gp in self.guardpath:
            if os.path.commonprefix([gp, modpath]) == gp:
                return True
        return False

    def _load_module(self):
        # add guardian paths to the base python search paths so that
        # the module can import other guardian modules or user code
        # with the standard import mechanism
        # FIXME: is there a cleaner way to do this without adding to
        # the overall sys.path?  do we care?
        if self.guardpath not in sys.path:
            sys.path += self.guardpath

        # add IFO and SYSTEM as global variables
        # FIXME: this has to be added to the builtins so that it's
        # globally available in the module we're loading and in
        # anything that it imports.  This isn't clean since it's
        # obviously poluting the global namespace.
        # FIXME: deprecate _*_ variables
        builtins.IFO = self.ifo
        builtins._IFO_ = self.ifo
        builtins.SYSTEM = self.name
        builtins._SYSTEM_ = self.name

        # FIXME: make system module variables available to any other
        # guardian modules it imports

        # FIXME: module reload does not support REMOVAL of states

        # variable to store path to any module loaded from guardpath
        # self._import will populate this as it imports new modules
        # from guardpath
        self._usercode = []

        # reload only once a given load cycle.  The defaultdict is
        # False by default.  When we come to a module we intend to
        # reload, note it in the dict.  That way we don't keep
        # reloading it for every module that imports it.
        self.__already_loaded = defaultdict(bool)

        # a little import chicanery here: override the import builtin
        # during the loading of system module with our custom import
        # runction that reloads any module found in guardpath.  we
        # need to do this so that any user code imported by a system
        # module will also be reloaded.
        # FIXME: there must be a cleaner way to do this.
        builtins.__import__ = self._import
        self._module = self._import(self._modname)
        builtins.__import__ = _builtin__import__

        return self._module

    def add_state(self, name, obj):
        """Add a state to the system.

        """
        if networkx.is_frozen(self._graph):
            raise GuardSystemLoadError("system is frozen.")

        if name in [i[0] for i in const.RESERVED_STATES]:
            raise GuardSystemLoadError("'%s' is a reserved state name." % (name))

        if len(name) > const.STATE_NAME_LENGTH:
            raise GuardSystemLoadError("state name over %d character limit: %s (%d characters)" % (const.STATE_NAME_LENGTH, name, len(name)))

        if not isinstance(obj.goto, bool) and not isinstance(obj.goto, int):
            raise GuardSystemLoadError("state 'goto' attribute must be a bool or int: %s" % (name))

        if not isinstance(obj.request, bool):
            raise GuardSystemLoadError("state 'request' attribute must be a bool: %s" % (name))

        if not isinstance(obj.redirect, bool):
            raise GuardSystemLoadError("state 'redirect' attribute must be a bool: %s" % (name))

        if hasattr(obj, 'index'):
            index = obj.index

            if not isinstance(index, int):
                raise GuardSystemLoadError("state 'index' attribute must be an int: %s" % (name))

            if name == 'INIT' and index != 0:
                raise GuardSystemLoadError("INIT state must be index 0 (assigned 0 if unspecified).")

            if name != 'INIT' and index <= 0:
                raise GuardSystemLoadError("state index must be positive definite: %s" % (name))

            if index in self.indices and name != self.index(index):
                raise GuardSystemLoadError("state index %d already in use by state %s: %s" % (index, self.index(index), name))

        else:
            if name == 'INIT':
                index = 0
            else:
                # assign first free negative integer as index,
                # excluding a set of reserved indices. we use negative
                # numbers to indicate that the index is auto-assigned.
                index = min(self.indices + [i[1] for i in const.RESERVED_STATES]) - 1

        # check that state is valid by attempting to instantiate it
        o = obj()

        # finally, add the state to the graph
        self._graph.add_node(
            name,
            code=obj,
            goto=obj.goto,
            request=obj.request,
            redirect=obj.redirect,
            index=index,
        )

    def add_edge(self, edge, goto=False):
        """Add an edge to the system.

        Edges are tuples of states of the form (SOURCE, DESTINATION).
        An integer edge weight may be specified as the third element
        of the tuple.  If not specified, edges have a weight of 1.

        """
        if networkx.is_frozen(self._graph):
            raise GuardSystemLoadError("system is frozen.")

        data = {}
        data['goto'] = goto
        data['weight'] = 1
        if len(edge) > 2:
            if type(edge[2]) != int:
                raise GuardSystemLoadError("edge weights must be integers: %s" % str(edge))
            data['weight'] = edge[2]

        edge = edge[:2]
        for state in edge:
            if state not in self:
                raise GuardSystemLoadError("unknown edge state: %s" % state)

        # skip if we already have this edge
        if self._graph.has_edge(*edge):
            return
            raise GuardSystemLoadError("system already contains edge: %s" % str(edge))

        # finally, add the edge to the graph
        self._graph.add_edge(*edge, **data)

    def load(self):
        """Import system module and populate system state graph.

        If a module was not defined at initialization an exception
        will be thrown.

        Any previously defined system graph is overwritten before
        loading the module.  The system graph is frozen once the
        module has been fully loaded.

        """
        if not self._modpath:
            raise GuardSystemLoadError("No module specified for this system.")

        self._reset()

        # load the module object
        module = self._load_module()

        # set channel access prefix from module
        # only do this if prefix not specified on system invocation
        if self._ca_prefix is None:
            # use prefix specified in module
            if hasattr(module, 'ca_prefix'):
                self._ca_prefix = module.ca_prefix
            # FIXME: deprecate
            elif hasattr(module, 'prefix'):
                self._ca_prefix = module.prefix

        ###########################
        # add states as graph nodes

        # add default empty INIT state.  this state must always be
        # present, as it's the default start up state
        self.add_state('INIT', GuardState)

        # loop through objects in module and extract all GuardState
        # objects as nodes
        for key, obj in inspect.getmembers(module):

            # register manager objects
            if isinstance(obj, (Node, NodeManager)):
                # if a NodeManager has already been registered with this name,
                # we want to extract it's existing Node objects, since they
                # contain state.
                if key in self._node_managers:
                    mngr = self._node_managers[key]
                    for node in obj:
                        if node.name in mngr:
                            obj.nodes[node.name] = mngr[node]
                self._node_managers[key] = obj
                continue

            # reject everything else that's not a class definition
            if not inspect.isclass(obj):
                continue

            # retrieve all base classes of the object
            bases = inspect.getmro(obj)

            # extract on GuardState object
            if GuardState not in bases:
                continue
            # ignore the GuardState object itself (each object is
            # first in it's own mro list)
            if bases[0] is GuardState:
                continue

            # ignore states starting with underscore ('_') which are
            # used to indicate inheritable base states
            if key[0] == '_':
                continue

            self.add_state(key, obj)

        # set initial request
        if hasattr(module, 'request'):
            if module.request not in self:
                raise GuardSystemLoadError("unknown initial request state: %s" % module.request)
            if not self.is_request(module.request):
                raise GuardSystemLoadError("initial request state must be requestable (request=True): %s" % module.request)
            if module.request == 'INIT':
                raise GuardSystemLoadError("INIT may not be used as the initial request")
            self._initial_request = module.request
        else:
            self._initial_request = None

        # set nominal state
        if hasattr(module, 'nominal'):
            if module.nominal not in self:
                raise GuardSystemLoadError("unknown nominal state: %s" % module.nominal)
            self._nominal_state = module.nominal
        else:
            self._nominal_state = None

        # set manager
        if hasattr(module, 'GRD_MANAGER'):
            self._manager = module.GRD_MANAGER
        else:
            self._manager = None

        # add specified edges
        if hasattr(module, 'edges'):
            # verify states
            for edge in module.edges:
                self.add_edge(edge)

        # add edges from every state to every goto
        for goto in self.goto_states:
            if type(self[goto].goto) is int:
                weight = self[goto].goto
            else:
                weight = const.DEFAULT_GOTO_EDGE_WEIGHT
            for s in self:
                # skip if this is the same state
                if s == goto:
                    continue
                edge = (s, goto, weight)
                self.add_edge(edge, goto=True)

        ###########################
        # other system attributes

        if hasattr(module, 'ca_monitor'):
            if type(module.ca_monitor) is not bool:
                raise GuardSystemLoadError("'ca_monitor' setting must be a bool.")
            self._ca_monitor = module.ca_monitor
        if hasattr(module, 'ca_monitor_notify'):
            if type(module.ca_monitor_notify) is not bool:
                raise GuardSystemLoadError("'ca_monitor_notify' setting must be a bool.")
            self._ca_monitor_notify = module.ca_monitor_notify
        if hasattr(module, 'ca_channels'):
            for channel in module.ca_channels:
                if type(channel) not in [str, tuple]:
                    raise GuardSystemLoadError("'ca_channel' elements must be string or tuple.")
            self._ca_channels = module.ca_channels

        if hasattr(module, 'related_displays'):
            try:
                module.related_displays[0][2]
            except:
                raise GuardSystemLoadError("\"related_display\" attribute must be a list of (label,name,args) tuples.")
            self._related_displays = module.related_displays

        # freeze the graph to prevent modification after creation
        networkx.freeze(self._graph)

        return self

    ##########

    @property
    def guardpath(self):
        """Module search path."""
        return self._guardpath

    @property
    def ifo(self):
        """System IFO."""
        return self._ifo

    @property
    def name(self):
        """System name."""
        return self._name

    @property
    def modname(self):
        """Name of module object."""
        return self._modname

    @property
    def path(self):
        """Primary module path."""
        return self._modpath

    @property
    def usercode(self):
        """List of "usercode" modules loaded from guardpath."""
        return self._usercode

    @property
    def module(self):
        """Loaded module object."""
        return self._module

    @property
    def ca_prefix(self):
        """EPICS channel access prefix."""
        return self._ca_prefix

    @property
    def ca_monitor(self):
        """EPICS setting monitor bool."""
        return self._ca_monitor

    @property
    def ca_monitor_notify(self):
        """EPICS setting monitor notification bool."""
        return self._ca_monitor_notify

    @property
    def ca_channels(self):
        """EPICS setting monitor channel initializations."""
        return self._ca_channels

    @property
    def related_displays(self):
        """User code specified MEDM related displays.

        List of (label, name, args) string tuples.
        """
        return self._related_displays

    @property
    def request(self):
        """User code specified REQUEST state."""
        return self._initial_request

    @property
    def nominal(self):
        """User code specified NOMINAL state."""
        return self._nominal_state

    @property
    def manager(self):
        """User code specified MANAGER."""
        return self._manager

    @property
    def graph(self):
        """System state graph object."""
        return self._graph

    def is_goto(self, state):
        """Return True if state is a goto state."""
        return self._graph.nodes[state]['goto']

    def is_request(self, state):
        """Return True if state is a requestable state."""
        return self._graph.nodes[state]['request']

    def can_redirect(self, state):
        """Return True if state is redirectable."""
        return self._graph.nodes[state]['redirect']


    def states_iter(self, reverse=True, filters=[], init_last=True):
        """Generator of (state, index, requestable) tuple.

        States are ordered by descending index by default.  Set
        'reverse=False' to order in ascending index.

        'filters' is a list of tuples of state data (key, value) pairs
        such that the returned states will have data[key] == value.

        'init_last' sets whether or not the INIT state is appended
        separately at the end of the list, or inserted in index order.

        """
        def filt(filters, data):
            r = True
            for filt in filters:
                r &= data[filt[0]] == filt[1]
            return r
        for state, data in sorted(self._graph.nodes(data=True),
                                  key=lambda out: out[1]['index'],
                                  reverse=reverse):
            if init_last and state == 'INIT':
                continue
            if filt(filters, data):
                yield (state, data['index'], data['request'])
        if init_last and 'INIT' in self._graph:
            state = 'INIT'
            data = self._graph.nodes[state]
            if filt(filters, data):
                yield (state, data['index'], data['request'])

    @property
    def states(self):
        """Ordered list of state names."""
        return [data[0] for data in self.states_iter()]

    @property
    def goto_states(self):
        """Ordered list of goto state names."""
        return [data[0] for data in self.states_iter(filters=[('goto', True)])]

    @property
    def request_states(self):
        """Ordered list of requestable state names."""
        return [data[0] for data in self.states_iter(filters=[('request', True)])]


    @property
    def edges(self):
        """List of graph directed edge tuples."""
        return self._graph.edges()

    ##########

    @property
    def node_managers(self):
        """List of NodeManager objects"""
        return self._node_managers.values()

    ##########

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.name)

    def __str__(self):
        return "<%s %r>" % (self.__class__.__name__, self.name)

    def __contains__(self, state):
        """Return True if state in system."""
        return state in self._graph

    def __len__(self):
        """Return number of states in system."""
        return len(self._graph)

    def __iter__(self):
        """Iterator over state names."""
        return iter(self._graph.nodes)

    def _data_for_key(self, key):
        """Return state (key, data) tuple for state.

        If a string is provided, key is the state index.  If a number
        is provided, key is the state name.

        KeyError or TypeError exceptions are raised where appropriate.

        """
        if isinstance(key, str):
            try:
                data = self._graph.nodes[key]
                return data['index'], data
            except KeyError:
                raise KeyError("%s is not a state name" % key)

        elif isinstance(key, int):
            for state, index in self._graph.nodes(data='index'):
                if index == key:
                    return state, self._graph.nodes[state]
            raise KeyError("%s is not a state index" % key)

        else:
            raise TypeError("item must be state name string or index integer.")

    def __getitem__(self, key):
        """Get state object from state name string or index."""
        return self._data_for_key(key)[1]['code']

    ##########

    @property
    def indices(self):
        """List of state indices."""
        return [index for state, index in self._graph.nodes(data='index')]

    def state_index_dict(self):
        """Return a dict of state:index pairs."""
        return {state: index for state, index in self._graph.nodes(data='index')}

    def index_state_dict(self):
        """Return a dict of index:state pairs."""
        return {index: state for state, index in self._graph.nodes(data='index')}

    def index(self, key):
        """Convert state name to index or vice versa.

        If a string is provided it is interpreted as a state name and
        the corresponding state index is returned.  If an number is
        provided it is interpreted as a state index and the
        corresponding state name string is returned.

        """
        return self._data_for_key(key)[0]

    ##########

    def shortest_path(self, source, request):
        """Return shortest path between *source* and *request* as list of state names.

        When multiple paths exist between *source* and *request*, the
        shortest path is the one with the lowest edge weight sum.  By
        default, edges have a weight of one.  Edge weights may be
        positive or negative.  If multiple equal weight path exists,
        one will be chosen arbitrarily.

        If the *source* is in the state graph the path will be
        calculated directly.  If it is not (source == None), traverse
        the graph looking for a goto state closest to the requested
        state, and use that as the source.

        If there is no path between the source and the request, return
        None.

        """
        # check validity of state and request arguments
        if source:
            self[source]
        self[request]

        # if the source in not specified (None), find the goto with
        # the shortest path to the request
        if not source and self.goto_states:
            slen = 0
            sg = None
            for g in self.goto_states:
                if networkx.has_path(self.graph, g, request):
                    ilen = 1./len(networkx.shortest_path(self.graph, g, request))
                    if ilen > slen:
                        sg = g
            if sg:
                source = sg

        if not networkx.has_path(self.graph, source, request):
            return None

        return networkx.shortest_path(self.graph, source, request, weight='weight')



from __future__ import print_function
import os
import sys
import json
import signal
import logging
import argparse
import functools
import subprocess
try:
    from termcolor import cprint
except ImportError:
    def cprint(*args):
        print(args[0])
from datetime import timedelta
from gpstime import gpstime, GPSTimeException
from dateutil.tz import tzutc, tzlocal

from . import systemd
from . import svlogd
from . import util

##################################################

PROG = 'guardctrl'

parser = argparse.ArgumentParser(
    prog=PROG,
    description="""Guardian daemon supervision control interface.

Control guardian daemon processes managed by the site systemd
supervision system.
""",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog="""
Add '-h' after individual commands for command help.

Node names may be specified with wildcard/globbing, e.g. 'SUS_*'.
A single '*' will act on all configured nodes (where appropriate).
""",
)

parser.add_argument('-d', '--debug', action='store_true',
                    help="print debug information to stderr")
# parser.add_argument('-v', '--version', action='version',
#                     version='Guardctrl {}'.format(__version__),
#                     help="print guardctrl version and exit")

subparser = parser.add_subparsers(
    title='Commands',
    metavar='<command>',
    dest='cmd',
    #help=argparse.SUPPRESS,
)
subparser.required = True

def gen_subparser(cmd, func, node_nargs=None, **kwargs):
    help = func.__doc__.split('\n')[0].lower().strip('.')
    desc = func.__doc__
    p = subparser.add_parser(
        cmd,
        #formatter_class=argparse.RawDescriptionHelpFormatter,
        help=help,
        description=desc,
        **kwargs
    )
    p.set_defaults(func=func)
    if node_nargs:
        p.add_argument('nodes', metavar='node', nargs=node_nargs,
                       help="guardian node name (may be wildcard/glob pattern)")
    return p

##################################################

def command_help(args):
    """Show command help."""
    if args.cmd:
        try:
            p = subparser.choices[args.cmd]
        except KeyError:
            sys.exit("Unknown command: {}".format(args.cmd))
    else:
        p = parser
    p.print_help()

p = gen_subparser("help", command_help)
p.add_argument('cmd', metavar='<command>', nargs='?', help="command")


def list_nodes(args):
    """List nodes and node supervision state."""
    nodes = args.nodes
    include_status = args.include_status
    only_active = args.only_active
    if not nodes:
        nodes = '*'
    nodes = list(systemd.list_nodes(nodes))
    if not nodes:
        return
    if not include_status and not only_active:
        for node in nodes:
            print(node)
        return
    node_w = max([len(node) for node in nodes]) + 2
    if include_status:
        fmt = '{{node:<{node_w}}} {{enabled:<{enabled_w}}} {{active:<{active_w}}} {{execstart}}'.format(
            node_w=node_w,
            enabled_w=10,
            active_w=10,
        )
    else:
        fmt = '{node}'
    for node, enabled, active, execstart in systemd.list_nodes(
            nodes,
            only_active=only_active,
            include_status=True):
        color = None
        attrs = []
        if active == 'active':
            color = 'green'
        elif active == 'failed':
            color = 'red'
            attrs = ['bold']
        else:
            color = 'yellow'
        line = fmt.format(node=node, enabled=enabled, active=active, execstart=execstart)
        cprint(line, color, attrs=attrs)

p = gen_subparser('list', list_nodes, node_nargs='*')
p.add_argument('-s', '--status', dest='include_status', action='store_true',
               help="include node status columns")
p.add_argument('-r', '--running', dest='only_active', action='store_true',
               help="only list running (active) nodes")


@functools.wraps(systemd.print_node_status)
def print_node_status(args):
    nodes = args.nodes
    try:
        return systemd.print_node_status(nodes, print_logs=True)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('status', print_node_status, node_nargs='*')


@functools.wraps(systemd.enable_nodes)
def enable_nodes(args):
    nodes = args.nodes
    now = args.now
    try:
        return systemd.enable_nodes(nodes, now=now)
    except util.GuardCtrlError as e:
        sys.exit(e)

p = gen_subparser('enable', enable_nodes, node_nargs='+')
p.add_argument('-n', '--now', action='store_true',
               help="start nodes immediately")


@functools.wraps(systemd.start_nodes)
def start_nodes(args):
    nodes = args.nodes
    try:
        return systemd.start_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('start', start_nodes, node_nargs='+')


@functools.wraps(systemd.restart_nodes)
def restart_nodes(args):
    nodes = args.nodes
    try:
        return systemd.restart_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('restart', restart_nodes, node_nargs='+')


@functools.wraps(systemd.stop_nodes)
def stop_nodes(args):
    nodes = args.nodes
    try:
        return systemd.stop_nodes(nodes)
    except util.GuardCtrlError as e:
        sys.exit(e)

gen_subparser('stop', stop_nodes, node_nargs='+')


@functools.wraps(systemd.disable_nodes)
def disable_nodes(args):
    nodes = args.nodes
    now = args.now
    node_list = list(systemd.list_nodes(nodes))
    if not node_list:
        sys.exit("Unknown nodes: {}".format(' '.join(nodes)))
    print("Really disable the following nodes?", file=sys.stderr)
    for node in node_list:
        print("  {}".format(node), file=sys.stderr)
    print("This will remove these nodes from the site list.", file=sys.stderr)
    ret = input("Type 'yes' to disable: ")
    if ret != 'yes':
        sys.exit('Aborted.')
    try:
        return systemd.disable_nodes(node_list, now=now)
    except util.GuardCtrlError as e:
        sys.exit(e)

p = gen_subparser('disable', disable_nodes, node_nargs='+')
p.add_argument('-n', '--now', action='store_true',
               help="stop nodes immediately")

##################################################

def datetime(ds):
    """wrapper function to parse datetime strings to gpstime objects.

    Throws TypeError needed for argparse type checking.

    """
    try:
        return gpstime.parse(ds)
    except GPSTimeException as e:
        raise TypeError(e)


def print_logs(args):
    """View node logs.

Logs will be followed by default, unless the +f flag is specified.
Log timestamps are in UTC by default.
    """
    node = args.nodes
    time_format = args.time_format
    search_old_logs = args.search_old_logs
    # ddelta = timedelta(seconds=float(args.duration))

    # if carg('hours'):
    #     args.after = gpstime.now(tz=tzlocal()) - timedelta(seconds=3600*float(args.hours))
    # elif carg('after'):
    #     args.after = _parse_time_arg('after')
    #     if carg('before'):
    #         args.before = _parse_time_arg('before')
    #         if args.before < args.after:
    #             raise util.GuardCtrlError("Before time is earlier than after time.")
    #     elif carg('duration'):
    #         args.before = args.after + ddelta
    # elif carg('center'):
    #     dur = ddelta/2
    #     args.after = _parse_time_arg('center') - dur
    #     args.before = _parse_time_arg('center') + dur
    # elif carg('before'):
    #     args.before = _parse_time_arg('before')
    #     args.after = args.before - ddelta
    # # else:
    # #     args.follow = True

    logging.debug("search after : %s" % args.after)
    logging.debug("search before: %s" % args.before)

    if time_format == 'utc':
        def dt2s(dt):
            return dt.astimezone(tzutc()).strftime(util.ISO_FMT+'Z')
    elif time_format == 'local':
        def dt2s(dt):
            return dt.strftime(util.ISO_FMT)
    elif time_format == 'gps':
        def dt2s(dt):
            return '{:f}'.format(gpstime.fromdatetime(dt).gps())

    if search_old_logs:
        kwargs = {k: getattr(args, k) for k in ['nodes', 'after', 'before']}
        gen = svlogd.node_log_search(**kwargs)
    else:
        kwargs = {k: getattr(args, k) for k in ['nodes', 'after', 'before', 'nlines', 'follow']}
        gen = systemd.node_logs(**kwargs)

    for dt, msg in gen:
        if dt:
            print('{} {}'.format(dt2s(dt), msg))
        else:
            print(msg)

p = gen_subparser('log', print_logs, node_nargs='*', prefix_chars='-+')
p.epilog = """
Date/time strings for filters are handled by a date/time string parser
and therefore may be specified in various formats.  If --after and
--before are not both specified, --duration will be used.  --lines
will ultimately limit the number of lines returned.

Examples:

  Show all ISC_LOCK logs from Sept. 14, 2015:

  $ {prog} --after 2015-09-14 --before 2015-09-15 ISC_LOCK

  Show the last 1000 lines of all SUS node logs:

  $ {prog} -n1000 SUS_*

  Quote compound date strings, and specify 'UTC' for times in UTC:

  $ {prog} -a "2016-08-20 06:23:22 UTC" -b "2016-08-20 06:23:52 UTC" ISC_LOCK

  GPS times are accepted as well:

  $ {prog} 1155709419 ISC_DRMI

""".format(prog=PROG+' log')
p.formatter_class = argparse.RawDescriptionHelpFormatter
p.add_argument('-f', '--follow', action='store_true',
               help="follow logs")
p.add_argument('-a', '--after', metavar='DATETIME', type=datetime,
               help="filter for logs after specified date/time")
p.add_argument('-b', '--before', metavar='DATETIME', type=datetime,
               help="filter for logs before specified date/time")
# p.add_argument('-c', '--center', metavar='DATETIME', type=datetime,
#                help="filter for logs centered around specified time")
# p.add_argument('-o', '--hours', 'hours', type=int, metavar='HOURS',
#                help="filter for logs from the last H hours")
# p.add_argument('-t', '--day', 'hours', flag_value=24,
#                help="filter for logs from the last 24 hours")
# p.add_argument('-d', '--duration', metavar='SECONDS', type=float, default=60)
p.add_argument('-n', '--lines', metavar='N', dest='nlines', type=int, default=100,
               help="limit logs to N lines [100]")
p.add_argument('-u', '--utc', dest='time_format', action='store_const', const='utc',
               help="show log lines with UTC timestamps (default)")
p.add_argument('-l', '--local', dest='time_format', action='store_const', const='local',
               help="show log lines with local timestamps")
p.add_argument('-g', '--gps', dest='time_format', action='store_const', const='gps',
               help="show log lines with GPS timestamps")
p.set_defaults(time_format='utc')
p.add_argument('--old', dest='search_old_logs', action='store_true',
               help="search old logs (<=O2, some arguments ignored)")

##################################################

def system_command(cmd, argv):
    cmd = [cmd] + argv
    logging.debug("cmd: {}".format(' '.join(cmd)))
    try:
        sys.exit(subprocess.run(cmd).returncode)
    except KeyboardInterrupt:
        sys.exit()

subparser.add_parser('env')
subparser.add_parser('systemctl')
subparser.add_parser('journalctl')

##################################################

def main_local():
    # handle invocation as ssh ForceCommand
    ssh_args = os.getenv('SSH_ORIGINAL_COMMAND')
    if ssh_args and len(sys.argv) == 1:
        # assume json argument encoding if the first character is the
        # json bracket list start character
        if ssh_args[0] == '[':
            sys.argv += json.loads(ssh_args)
        else:
            sys.argv += os.getenv('SSH_ORIGINAL_COMMAND').split()

    # parse known args first, just to get debug flag and handle direct
    # sys commands
    args, remainder = parser.parse_known_args()

    log_level = os.getenv('LOG_LEVEL', 'INFO').upper()
    if args.debug:
        log_level = 'DEBUG'
    FORMAT = '%(levelname)s: %(module)s: %(message)s'
    logging.basicConfig(format=FORMAT, level=log_level)

    logging.debug(sys.argv)

    # handle direct systemctl/journalctl invocation
    if args.cmd in ['env', 'systemctl', 'journalctl']:
        system_command(args.cmd, remainder)

    else:
        # re-parse all args to properly fail for unknown args
        args = parser.parse_args()
        args.func(args)


def main_remote():
    parser = argparse.ArgumentParser(prefix_chars='-+', add_help=False)
    parser.add_argument('+t', dest='terminal', action='store_false')
    wargs, args = parser.parse_known_args()

    IFO = os.getenv('IFO')
    GUARDCTRL_USER = os.getenv('GUARDCTRL_USER', 'guardian')
    GUARDCTRL_HOST = os.getenv('GUARDCTRL_HOST', '{}guardian'.format(IFO.lower()))

    SSH_CMD = [
        'ssh',
        '-S', '~/.ssh/controls/%r@%h:%p',
        '-o', 'ControlMaster=auto',
        '-o', 'ControlPersist=600',
    ]

    # force TTY by default, with "+t" option to unset and allow
    # separate stdout/stderr streams
    if wargs.terminal:
        SSH_CMD += ['-t']

    SSH_CMD += [
        '{}@{}'.format(GUARDCTRL_USER, GUARDCTRL_HOST),
        '--',
    ]

    # create a directory for the ssh control sockets
    try:
        os.makedirs(os.path.expanduser('~/.ssh/controls/'))
    except:
        pass

    # json encode commands for ssh, to pass spaces in a reasonable way
    args_enc = json.dumps(args)

    # launch in foreground
    try:
        subprocess.run(
            SSH_CMD + [args_enc],
            stderr=subprocess.PIPE,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        sys.exit("cmd: {}\n{}".format(
            ' '.join(e.cmd), e.stderr.decode().strip()))


def main():
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    if os.path.exists(os.path.expanduser('~/.guardctrl-home')) \
       or os.getenv('GUARDCTRL_LOCAL'):
        main_local()

    elif os.getenv('GUARDCTRL_HOST'):
        main_remote()

    else:
        main_local()


if __name__ == '__main__':
    main()



# -*- fill-column: 80; -*-

import os
import time
import logging
import traceback
try:
    from setproctitle import setproctitle
except ImportError:
    def setproctitle(*args):
        pass
try:
    from systemd.daemon import notify as sd_notify
except ImportError:
    def sd_notify(*args):
        pass

############################################################

from ._version import __version__
from . import const
from . import db
from .worker import Worker

############################################################

class GuardDaemonError(Exception):
    pass

class Daemon(object):
    """Guardian daemon.

    """
    def __init__(self, system, logger,
                 initial_op='EXEC',
                 initial_mode=None,
                 initial_state=None,
                 initial_request=None,
                 single_shot=False,
                 archive=None,
                 ):
        # check arguments
        if initial_op not in db.DAEMON_OP:
            raise GuardDaemonError("Initial OP not in %s" % str(DAEMON_OP))
        if initial_mode and initial_mode not in db.DAEMON_MODE:
            raise GuardDaemonError("Initial MODE ({}) not in {}".format(initial_mode, str(DAEMON_MODE)))

        # FIXME: check that system is initialized already
        self.system = system
        self.log = logger
        self.redirect_timeout = const.REDIRECT_TIMEOUT
        self.single_shot = single_shot

        # set the default loglevel to be the initialized loglevel
        self.default_loglevel = logging.getLevelName(self.log.getEffectiveLevel())

        self.log.info("Guardian v%s", __version__)

        if self.single_shot:
            self.log.info("single-shot mode")

        # worker subprocess
        self.worker = None

        # store redirect request time
        # FIXME: this is the only internal state variable that is not
        # an exposed PV.  Should it be?
        self.redirect_request_time = None

        # EPICS channel access server thread
        if self.single_shot:
            self.cas = None
        else:
            from .cas import CAServer
            self.cas = CAServer(self.system.name)
            self.log.info("EPICS control prefix: %s", self.cas.prefix)

        # archive git repo
        if archive:
            self.archive = archive
            self.log.info("system archive: %s" % (self.archive.root))
        else:
            self.archive = None

        # initialize internal db
        self.db = db.Database(self.cas)

        # initialize state index and request enum
        self.db.update_system(self.system, init=True)

        # set initial values
        self['TIME_INIT'] = int(time.time())
        self['OP'] = initial_op
        if initial_mode is not None:
            self['MODE'] = initial_mode
        elif self.system.manager:
            self['MANAGER'] = self.system.manager
            self['MODE'] = 'MANAGED'
        else:
            self['MODE'] = 'AUTO'
        self['LOGLEVEL'] = self.default_loglevel
        self['STATE'] = initial_state or 'INIT'
        self['TARGET'] = self['STATE']
        self['REQUEST'] = initial_request or self.system.request or 'NONE'
        self['NOMINAL'] = self.system.nominal or 'NONE'
        self['SPM_MONITOR'] = self.system.ca_monitor
        self['SPM_MONITOR_NOTIFY'] = self.system.ca_monitor_notify
        self['STATUS'] = 'INIT'

        self.archive_commit('DAEMON INIT')
        self.archive_print()

        self.log.info("system name: %s", self.system.name)
        self.log.info("system CA prefix: %s", self.system.ca_prefix)
        self.log.info("module path: %s", self.system.path)
        if self.system.usercode:
            for code in self.system.usercode:
                self.log.info("user code: %s", code)
        self.log.debug("system states: %s", self.system.states)
        self.log.debug("request states: %s", self.system.request_states)
        self.log.info("initial state: %s" % self['STATE'])
        self.log.info("initial request: %s" % self['REQUEST'])
        if not self.single_shot:
            self.log.info("nominal state: %s" % self['NOMINAL'])
        if self.system.manager:
            self.log.info("initial manager: %s" % self['MANAGER'])
        self.log.info("CA setpoint monitor: %s" % self['SPM_MONITOR'])
        self.log.info("CA setpoint monitor notify: %s" % self['SPM_MONITOR_NOTIFY'])

        self.log.info("daemon initialized")

    def single_shot_exit(self, code=0, message=None):
        if self.single_shot:
            if code != 0 and message:
                self.log.error(message)
            raise SystemExit(code)

    ########################################
    # archive

    def archive_print(self):
        if not self.archive:
            return
        hexid = self.archive.get_hexsha()
        intid = self.archive.get_intsha()
        self.log.info("system archive: id: %s (%d)" % (hexid, intid))

    def archive_commit(self, message):
        if not self.archive:
            return
        message += '''

guardian version: {}
'''.format(__version__)
        ret = False
        with self.archive as archive:
            if archive.commit(self.system, message):
                self.log.info("system archive: code changes detected and committed")
                ret = True
        self['ARCHIVE_ID'] = self.archive.get_intsha()
        return ret

    ########################################
    # "magic" methods for accessing/setting the process variables

    def __getitem__(self, channel):
        return self.db[channel]

    def __setitem__(self, channel, value):
        if self.db[channel] != value \
           and db.guarddb[channel].get('log_changes', True):
            self.log.debug('%s => %s' % (channel, value))
        self.db[channel] = value

    ########################################

    def message(self, message):
        self.log.info(message)
        self['GRDMSG'] = message

    ########################################

    def worker_init(self):
        self.log.debug("initializing worker...")
        snapshot_file = None
        if self.archive:
            snapshot_file = os.path.join(self.archive.root, '%s.spm' % self.system.name)
        self.worker = Worker(self.system, self.log, snapshot_file)
        self.worker.start()
        self.log.info("worker ready")

    def worker_terminate(self):
        if self.worker is not None and self.worker.is_alive():
            self.log.debug("terminating worker...")
            self.worker.terminate()
            self.worker.join()
            self.log.info("worker terminated")
        self.worker = None

    ########################################
    ########################################

    def unstall(self):
        if self['STALLED']:
            self.log.info("STALL cleared")
            self['STALLED'] = False

    def raise_redirect(self):
        self['STATUS'] = 'REDIRECT'
        self.message("REDIRECT requested, timeout in %.3f seconds" % (self.redirect_timeout))
        self.redirect_request_time = time.time()

    def update_target(self, last_request):
        # this function updates the TARGET state based on the current
        # STATE and REQUEST.  It also potentially updates STATUS and
        # ERROR as needed.

        # in MANUAL mode, simply set target and redirect immediately
        if self['MODE'] == 'MANUAL' and self['REQUEST'] != 'NONE':
            self['TARGET'] = self['REQUEST']
            # FIXME: option to not redirect here, to let state finish?
            self.raise_redirect()
            return

        # request confirmations
        if self['REQUEST'] != last_request:
            # special INIT request handling: for INIT requests,
            # redirect to INIT immediately but reset the request back
            # to last request.
            if self['REQUEST'] == 'INIT':
                self.message("INIT redirect")
                self['REQUEST'] = last_request
                self['TARGET'] = 'INIT'
                self.raise_redirect()
                return

            # if changing from NONE requests to a request that
            # corresponds to the current state don't raise the normal
            # state re-request redirect.
            if self['REQUEST'] == self['STATE'] and last_request == 'NONE':
                self.log.info("request set from NONE to current state ({}); doing nothing.".format(self['STATE']))
                return

        # if the request is NONE, do nothing
        if self['REQUEST'] == 'NONE':
            # FIXME: if a NONE request comes in during a
            # non-requestable state, the state will continually be run
            # in the DONE status, even though that might not be
            # desirable
            self['TARGET'] = self['STATE']
            return

        # calculate shortest path from current state to current request.
        self.log.info("calculating path: %s->%s" % (self['STATE'], self['REQUEST']))
        newpath = self.system.shortest_path(self['STATE'], self['REQUEST'])

        # if path is None, there is no path so clean up and return
        if newpath is None:
            self.message("no path from %s->%s" % (self['STATE'], self['REQUEST']))
            self['TARGET'] = self['STATE']
            self.single_shot_exit(5)
            return

        self.log.debug("calculated path: %s" % newpath)

        # the second element of the path list is the target, or the
        # first element if the current state is the target.  if it's
        # different than the current target update the target
        if newpath[:2][-1] != self['TARGET']:
            self['TARGET'] = newpath[:2][-1]
            self.log.info("new target: %s" % self['TARGET'])

        # The following deals with state REDIRECTS.  a redirect occurs
        # when a new request results in a goto target state.  in this
        # case we don't wait for the current state to complete and
        # instead move directly to the goto state.  we also use a
        # redirect when the state equals the request and the user
        # requests the current state to be rerun.  we *don't* use a
        # redirect when there is a jump to a state with a goto target,
        # or the current state is INIT, as we want to make sure those
        # states are fully executed first before transitioning.
        #
        # a timeout in seconds is used to indicate that the worker
        # process should be forcibly killed if it has not returned in
        # the specified amount of time. this is harsh, since
        # terminating the worker destroys the worker process and the
        # existing EPICS (ezca) IO object, requiring the worker to be
        # restarted and all EPICS connections to be reestablished.

        # no redirects if we're stalled
        if self['STALLED']:
            return

        # first describe the in-coming and out-going edges to make the
        # logic clearer (True indicates a normal edge transition)
        #
        # inedge: False, True, 'jump'
        if self['STATUS'] == 'EDGE':
            inedge = True
        elif self['STATUS'] == 'JUMP':
            inedge = 'jump'
        else:
            inedge = False
        # outedge: False, True, 'goto'
        if self['TARGET'] == self['STATE']:
            outedge = False
        elif self.system.is_goto(self['TARGET']) and self['STATE'] != 'INIT':
            outedge = 'goto'
        else:
            outedge = True
        # now handle the redirect condition itself
        if outedge == 'goto':
            if not self.system.can_redirect(self['STATE']):
                self.message("GOTO REDIRECT IGNORED: redirect=False for state %s" % self['STATE'])
            else:
                self.message("GOTO REDIRECT")
                self.raise_redirect()
        elif self.single_shot:
            pass
        # same-state redirect only if state has edge to itself
        elif inedge == outedge == False and self.system.graph.has_edge(self['STATE'], self['STATE']):
            self.log.info("same state request redirect")
            self.raise_redirect()

    def reload_system(self):
        self.message("RELOAD requested.  reloading system data...")
        self.system.load()
        request_list_change = self.db.update_system(self.system)
        self.log.info("module path: %s", self.system.path)
        if self.system.usercode:
            for code in self.system.usercode:
                self.log.info("user code: %s", code)
        self.log.debug("system states: %s", self.system.states)
        self.log.debug("request states: %s", self.system.request_states)
        newnom = self.system.nominal or 'NONE'
        if newnom != self['NOMINAL']:
            self['NOMINAL'] = newnom
            self.log.info("new nominal state: %s" % self['NOMINAL'])
        if self.system.ca_monitor != self['SPM_MONITOR']:
            self['SPM_MONITOR'] = self.system.ca_monitor
            self.log.info("CA setpoint monitor: %s" % self['SPM_MONITOR'])
        if self.system.ca_monitor_notify != self['SPM_MONITOR_NOTIFY']:
            self['SPM_MONITOR_NOTIFY'] = self.system.ca_monitor_notify
            self.log.info("CA setpoint monitor notify: %s" % self['SPM_MONITOR_NOTIFY'])
        if self.archive_commit('DAEMON RELOAD'):
            self.archive_print()
        if request_list_change:
            self.message("RELOAD complete; REQUEST LIST CHANGED: *RELOAD MEDM CONTROL SCREEN*")
        else:
            self.message("RELOAD complete")

    ########################################
    ########################################

    def stop(self):
        # FIXME: what actions do we need to do here?
        sd_notify("STOPPING=1")
        self.log.info("stopping daemon...")
        self.worker_terminate()
        self.log.info("daemon stopped.")

    ########################################

    def run(self):
        self.log.info("============= daemon start =============")

        setproctitle('guardian {} {}'.format(self.system.name, self.system.path))

        # internal flags for noting transitions
        op = None
        mode = None
        manager = ''
        request = self['REQUEST']

        # holds time at which to reset log level
        loglevel_reset_time = None

        # handle times in int usec, to prevent cumulative round off error
        time_step = int(1000000/const.CPS)
        # start on the next integer second
        next_step = int(time.time() + 1) * 1000000 - time_step

        # start the worker subprocess
        self.worker_init()

        self.log.info("========== executing run loop ==========")
        loopstart = time.time()
        sd_notify("READY=1")

        ########################################
        ########################################
        while True:

            self['MAINUSEC'] = int((time.time() - loopstart)*1000000)

            # FIXME: this would be the best place to call this, as this should
            # only need to be called once to update all the values pushed during
            # the last part of the loop, as opposed to in the process loop
            # below, but see comment in cas.updatePVs.
            #if self.cas:
            #    self.cas.updatePVs()

            # sleep/process till next cycle
            next_step += time_step
            while int(time.time()*1000000) < next_step:
                if self.cas:
                    # NOTE: See comment in cas.upatePVs().  The best place I've
                    # found to put it is here, right before the server process
                    # happens, although I can't say exactly why.  If called too
                    # frequently (i.e. after every variable set or with the
                    # sleep time of this loop set too low) the daemon process
                    # will be in R more than S, jacking the system load up
                    # considerably.  If we put it here but set the process time
                    # too high, there's more skew in the total loop processing
                    # time.
                    self.cas.updatePVs()

                    # The main part of the loop below usually completely in less
                    # than 5ms, so there should be ~58ms in this process loop.
                    # A 5ms sleep here seems to be a reasonable compromise with
                    # the updatePVs constraints described above.  (Could adjust
                    # time as it gets closer.)
                    self.cas.process(0.005)
                else:
                    time.sleep(0.001)

            self.log.log(5, '====== start cycle ======')
            looplast = loopstart
            loopstart = time.time()
            self['LOOPUSEC'] = int((loopstart - looplast)*1000000)
            sd_notify("WATCHDOG=1")

            ########################################
            # bookeeping to be done every cycle

            # update heartbeat
            self['TIME_UP'] = int(time.time() - self['TIME_INIT'])

            # update the database to pull in any CA changes
            self.db.update()

            # note OP and MODE changes
            if self['OP'] != op:
                self.message("OP: %s" % self['OP'])
                op = self['OP']
            if self['MODE'] != mode:
                self.message("MODE: %s" % self['MODE'])
                mode = self['MODE']

            # check for worker failures
            if self.worker is not None and self.worker.exitcode is not None:
                raise GuardDaemonError("worker exited unexpectedly, exit code: %d" % self.worker.exitcode)

            # twiddle the log level
            if self['LOGLEVEL'] != logging.getLevelName(self.log.getEffectiveLevel()):
                self.log.setLevel(self['LOGLEVEL'])
                self.worker['LOGLEVEL'] = self['LOGLEVEL']
                self.log.info('LOGLEVEL => %s' % logging.getLevelName(self.log.getEffectiveLevel()))
                # reset log level in 1 minute
                loglevel_reset_time = time.time() + 60
            # reset back to default level after timeout
            if loglevel_reset_time and time.time() >= loglevel_reset_time:
                self['LOGLEVEL'] = self.default_loglevel
                self.log.setLevel(self['LOGLEVEL'])
                self.worker['LOGLEVEL'] = self['LOGLEVEL']
                loglevel_reset_time = None

            # retrieve status from worker
            if self.worker:
                # status
                self['WORKER'] = self.worker['STATUS']
                # notifications
                nmsgs = self.worker['NUSERMSGS']
                self['NOTIFICATION'] = nmsgs > 0
                # HACK: we use the "raw" value from the shmem
                # interface because of pcaspy bugginess when a char
                # array is set to an empty string.  The "raw" value is
                # actually an array of null bytes the same size as the
                # record, which seems to make things happier.
                for i in range(const.USERMSG_COUNT):
                    self['USERMSG'+str(i)] = self.worker._shmem['USERMSG'+str(i)].raw
                # floor to USERMSG_COUNT so 'ind' doesn't end up
                # greater than the number of USERMSG slots
                if nmsgs > const.USERMSG_COUNT:
                    nmsgs = const.USERMSG_COUNT
                if nmsgs > 0:
                    ind = str(int((time.time() / const.USERMSG_CYCLE_TIME) % nmsgs))
                    self['USERMSG'] = self.worker._shmem['USERMSG'+ind].raw
                else:
                    self['USERMSG'] = self.worker._shmem['USERMSG0'].raw
                # PV/setpoint monitoring
                self['PV_TOTAL'] = self.worker['PV_TOTAL']
                self['SPM_TOTAL'] = self.worker['SPM_TOTAL']
                self['SPM_CHANGED'] = self.worker['SPM_CHANGED']
                # HACK: see shmem hack note above
                for i in range(const.SPM_DIFF_COUNT):
                    self['SPM_DIFF'+str(i)] = self.worker._shmem['SPM_DIFF'+str(i)].raw
                    self['SPM_DIFF'+str(i)+'_S'] = self.worker._shmem['SPM_DIFF'+str(i)+'_S'].raw
                    self['SPM_DIFF'+str(i)+'_C'] = self.worker._shmem['SPM_DIFF'+str(i)+'_C'].raw
                    self['SPM_DIFF'+str(i)+'_D'] = self.worker._shmem['SPM_DIFF'+str(i)+'_D'].raw
                # PV snapshots
                if self['SPM_SNAP']:
                    self.worker.snapshot_request.set()
                    self['SPM_SNAP'] = False
                # manager status
                self['SUBNODES_TOTAL'] = self.worker['SUBNODES_TOTAL']
                self['SUBNODES_NOT_OK'] = self.worker['SUBNODES_NOT_OK']

            # log manager changes
            if self['MANAGER'] != manager:
                if self['MANAGER'] == '':
                    self.log.warning("MANAGER: removed")
                    self.unstall()
                    # FIXME: this should prompt a target update
                elif manager == '':
                    self.log.warning("MANAGER: assigned: %s", self['MANAGER'])
                else:
                    self.log.warning("MANAGER: reassigned, new manager: %s" % self['MANAGER'])
                manager = self['MANAGER']

            ##########
            # SUMMARY BITS

            self['ACTIVE'] = \
                self['OP'] == 'EXEC' \
                and \
                self['MODE'] in ['AUTO', 'MANAGED'] \
                and \
                self['LOAD_STATUS'] == 'DONE' \
                and \
                self['WORKER'] != 'INIT' \
                and \
                not self['ERROR'] \
                and \
                self['CONNECT'] == 'OK'

            self['READY'] = \
                self['ACTIVE'] \
                and \
                self['SUBNODES_NOT_OK'] == 0

            self['INTENT'] = \
                self['ACTIVE'] \
                and \
                self['REQUEST'] == self['NOMINAL'] \
                and \
                not self['STALLED']

            self['OK'] = \
                self['READY'] \
                and \
                self['INTENT'] \
                and \
                self['STATE'] == self['NOMINAL'] \
                and \
                self['STATUS'] == 'DONE'

            ########################################
            # handle code reload requests

            if self['LOAD']:
                self['LOAD_STATUS'] = 'REQUEST'
                self.message("LOAD REQUEST")
                self['LOAD'] = False
                try:
                    self.reload_system()
                    self.worker.load_request.set()
                    self['LOAD_STATUS'] = 'SET'
                    if self['ERROR']:
                        self['ERROR'] = False
                except:
                    self.log.error(traceback.format_exc())
                    self.message("LOAD ERROR: see log for more info (LOAD to reset)")
                    self['LOAD_STATUS'] = 'ERROR'
                    self['ERROR'] = True

            ########################################
            # transitons and target updates
            # but not if there's been a load error

            if not self['ERROR']:

                # update state on transition
                # status is updated only *after* path target is updated
                if self['STATUS'] in ['EDGE', 'JUMP']:
                    self.log.info("%s: %s->%s" % (self['STATUS'], self['STATE'], self['TARGET']))
                    self['STATE'] = self['TARGET']

                # calculate path and update target (this is dependent on the
                # details of the in and out edges)
                if self.db.request_event:
                    self.message("REQUEST: %s" % self['REQUEST'])
                    self.unstall()
                    self.update_target(request)
                    self.db.request_event = False
                    request = self['REQUEST']
                elif self['STATUS'] in ['INIT', 'EDGE', 'JUMP'] and self['MODE'] != 'MANUAL':
                    self.update_target(request)

                # status update only after target calculation
                if self['STATUS'] in ['INIT', 'EDGE', 'JUMP']:
                    self['STATUS'] = 'ENTER'

            ########################################
            ########################################
            ########################################

            self.log.log(5, '--------- %-4s  ---------' % self['OP'])

            # updatePVs here, roughly half way through the loop, so that the few
            # channels that change during both halfs of the loop will have their
            # all updates exposed.
            if self.cas:
                self.cas.updatePVs()

            ########################################
            # on STOP terminate the worker and restart it, reset the
            # state, and go to pause

            if self['OP'] == 'STOP':
                self['STATUS'] = 'INIT'
                self.worker_terminate()
                self.worker_init()
                if self['EXECTIME'] != 0:
                    self['EXECTIME_LAST'] = self['EXECTIME']
                self['OP'] = 'PAUSE'
                self['LOAD'] = True

            ########################################
            ########################################
            # each EXEC cycle attempt to acquire worker run lock.
            # acquiring it means the worker is currently idle,
            # allowing us to inspect the worker status and execute the
            # next function.

            if self.worker.acquire():
                self.log.log(5, "worker lock acquired")

                # update the latest worker status *after* acquiring
                # the worker lock
                self['WORKER'] = self.worker['STATUS']

                self.log.log(5, "exec status: %s", self['STATUS'])
                self.log.log(5, "worker status: %s" % self['WORKER'])

                # update "last" worker cycle time info
                # do this *before* the ERROR continue below
                if self.worker['EXECSTOP'] > self.worker['EXECSTART']:
                    self['EXECTIME'] = self.worker['EXECSTOP'] - self.worker['EXECSTART']
                    self['EXECTIME_LAST'] = self['EXECTIME']
                self['EXECCYCLES_LAST'] = self['EXECCYCLES']

                # abort any execution if we're in error
                if self['ERROR']:
                    self.worker.release()
                    self.log.log(5, "worker lock released")
                    continue

                # only reset the exec times if we're not in error
                self['EXECTIME'] = 0
                self['EXECCYCLES'] = 0

                # handle load transitions
                if self['LOAD_STATUS'] == 'SET':
                    self['LOAD_STATUS'] = 'INPROGRESS'
                elif self['LOAD_STATUS'] == 'INPROGRESS':
                    self['LOAD_STATUS'] = 'DONE'

                # clear connection errors if worker no longer
                # returning CERROR
                if self['CONNECT'] == 'ERROR' and self['WORKER'] != 'CERROR':
                    self.message("connections reestablished")
                    self['CONNECT'] = 'OK'

                ##########
                # EXECUTION EVALUTATION

                # if we're paused do nothing
                if self['OP'] == 'PAUSE':
                    pass

                # if we're entering a state, execute MAIN
                elif self['STATUS'] == 'ENTER':
                    self.message("executing state: %s (%s)" % (self['STATE'], self.system.index(self['STATE'])))
                    self['STATUS'] = 'MAIN'

                # if this is a redirect, transition to the redirect
                # target, or redo the current state
                elif self['STATUS'] == 'REDIRECT':
                    self.message("REDIRECT caught")
                    self.redirect_request_time = None
                    self.log.info("[%s.redirect]" % self['STATE'])
                    if self['TARGET'] == self['STATE']:
                        self['STATUS'] = 'ENTER'
                    elif self['MODE'] == 'MANUAL':
                        self['STATUS'] = 'JUMP'
                    else:
                        self['STATUS'] = 'EDGE'

                # otherwise evaluate the worker return status and make a decision
                elif self['STATUS'] in ['MAIN', 'RUN', 'DONE']:

                    # worker INIT indicates a worker restart, in which case we
                    # don't need to update the status
                    if self['WORKER'] == 'INIT':
                        pass

                    elif self['WORKER'] == 'CONTINUE':
                        self['STATUS'] = 'RUN'

                    elif self['WORKER'] == 'DONE':
                        if self['STATE'] == self['TARGET'] or self['STALLED']:
                            self['STATUS'] = 'DONE'
                            self.single_shot_exit(0, "[%s.exit]" % self['STATE'])
                        else:
                            self['STATUS'] = 'EDGE'

                    elif self['WORKER'] == 'JUMP':
                        self.message("JUMP target: %s" % self.worker['JUMP'])
                        if self['MODE'] == 'MANUAL':
                            self['STATUS'] = 'DONE'
                        else:
                            self.log.info("[%s.exit]" % self['STATE'])
                            self['STATUS'] = 'JUMP'
                            self['TARGET'] = self.worker['JUMP']
                        self.single_shot_exit(10)
                        # if MANAGED, jump transitions cause the node to STALL,
                        # meaning that EDGEs will not be followed until the
                        # stall is cleared by a REQUEST.
                        if self['MODE'] == 'MANAGED':
                            self.message('STALLED')
                            self['STALLED'] = True

                    elif self['WORKER'] == 'CERROR':
                        self.single_shot_exit(2, "EZCA CONNECTION ERROR, exiting.")
                        # do this check so we only log on new error
                        if self['CONNECT'] != 'ERROR':
                            self['CONNECT'] = 'ERROR'
                            self.message("EZCA CONNECTION ERROR. attempting to reestablish...")
                            self.log.error("CERROR: State method raised an EzcaConnectionError exception.")
                            self.log.error("CERROR: Current state method will be rerun until the connection error clears.")
                            self.log.error("CERROR: If CERROR does not clear, try setting OP:STOP to kill worker, followed by OP:EXEC to resume.")

                    elif self['WORKER'] == 'ERROR':
                        # if we're in the middle of a load caused by worker
                        # error, the main error would have been cleared during
                        # the initial reload, which would then drop us through
                        # to the execution phase here, even though the worker
                        # status error would still be present
                        if self['LOAD_STATUS'] == 'INPROGRESS':
                            pass

                        else:
                            self.message("ERROR in state %s: see log for more info (LOAD to reset)" % self['STATE'])
                            self['ERROR'] = True
                            self.single_shot_exit(1, "ERROR, exiting.")

                    # other worker return status ('COMMAND') should not occur
                    # and constitutes an internal error
                    else:
                        raise GuardDaemonError("Internal error: invalid WORKER in loop: %s" % self['WORKER'])

                # any other status ('INIT', 'EDGE', 'JUMP') should not occur and
                # constitutes an internal error
                else:
                    raise GuardDaemonError("Internal error: invalid STATUS in loop: %s" % self['STATUS'])

                ##########
                # execute method

                if self['ERROR']:
                    pass

                elif self['OP'] == 'PAUSE':
                    self.worker.command(self['STATE'], 'NOOP')

                elif self['STATUS'] == 'MAIN':
                    self.worker.command(self['STATE'], 'MAIN')

                elif self['STATUS'] in ['RUN', 'DONE']:
                    self.worker.command(self['STATE'], 'RUN')

                ##########
                # must release run lock for worker to acquire it,
                # or to reaquire it on the next cycle

                self.worker.release()
                self.log.log(5, "worker lock released")

            ########################################
            ########################################
            # if run lock can't be acquired, the worker is active and
            # still holding it, so do nothing unless we have a
            # redirect

            else:
                self.log.log(5, "worker locked")

                self['EXECTIME'] = time.time() - self.worker['EXECSTART']
                self['EXECCYCLES'] += 1

                # if we have a redirect, wait for the timeout then
                # forcibly terminate the worker and reset it.
                if self['STATUS'] == 'REDIRECT':
                    self.message("REDIRECT wait for worker completion...")
                    if time.time() >= self.redirect_request_time + self.redirect_timeout:
                        self.message("REDIRECT timeout reached. worker terminate and reset...")
                        self.worker_terminate()
                        self.worker_init()
                        self.redirect_request_time = None
                        self['STATUS'] = 'EDGE'



from __future__ import division
import math

from . import resources
from .screens import get_path

##################################################

MACROS = dict(
    IFO = '$(IFO)',
    SYSTEM = '$(SYSTEM)',
    SYSTEM_PATH = '$(SYSTEM_PATH)',
    )


def gen_macro_str(**kwargs):
    macros = MACROS.copy()
    if kwargs:
        macros.update(kwargs)
    return ','.join(['{}={}'.format(k, v) for k, v in macros.items()])


def gen_composite(**kwargs):
    path = get_path(kwargs['screen'])
    kwargs['macros'] = gen_macro_str(**kwargs.get('macros', {}))
    return '''
composite {{
        object {{
                x={x}
                y={y}
                width={width}
                height={height}
        }}
        "composite name"=""
        "composite file"="{path}; {macros}"
}}
'''.format(path=path, **kwargs)

##################################################

def write_screen_main(f, system):
    displays = system.related_displays + [('OVERVIEW',
                                           '$(USERAPPS)/sys/$(ifo)/medm/GUARD_OVERVIEW.adl',
                                           ''
                                           )]
    with open(get_path('GRD_MAIN.adl')) as fs:
        f.write(fs.read())
    f.write(resources.related_display_menu(displays,
                                           x=414,
                                           y=107,
                                           width=90,
                                           height=20,
                                           ))

def write_screen_compact(f):
    with open(get_path('GRD_COMPACT.adl')) as fs:
        f.write(fs.read())

def write_screen_minis(f):
    f.write(resources.header(bclr=14, height=400, width=400))
    y = 10
    for screen in ['GRD_NANO.adl', 'GRD_MICRO.adl', 'GRD_MINI.adl',
                   'GRD_MINI_USERMSG.adl', 'GRD_MINI_REQUEST_USERMSG.adl']: #, 'STATUSBAR']:
        f.write(gen_composite(screen=screen,
                              x=10, y=y, width=10, height=10))
        y += 70

def write_screen_states(f, system):
    states = system.states + ['NONE']
    nstates = len(states)

    margin = 10
    title_height = 18
    state_width = 305
    state_height = 25
    legend_height = 50

    # determine nice rectangle of state buttons
    sbar = state_width / float(state_height)
    rows = nstates
    cols = 1
    while rows > sbar*cols and nstates > 24:
        cols += 1
        rows = int(math.ceil(nstates/cols))
    def rc(i):
        col = int(math.floor(float(i)/rows))
        return i-col*rows, col

    offset = 84
    button_top = title_height + margin + offset
    width = margin + cols * state_width
    height = button_top + state_height*rows + margin
    if cols == 1:
        height += legend_height

    f.write(resources.header(height=height, width=width))
    f.write(resources.background(height=height, width=width))
    f.write(resources.title(x=48))

    f.write(gen_composite(screen='GRD_STATES_HEAD.adl',
                          x=10, y=title_height+margin, width=300, height=200))

    for i, state in enumerate(states):
        r,c = rc(i)
        x = margin + c*state_width
        y = button_top + r*state_height
        if state == 'NONE':
            index = -1
            is_goto = False
            is_request = True
        else:
            index = system.index(state)
            is_goto = system.is_goto(state)
            is_request = system.is_request(state)
        f.write(resources.state_request_button(x=x, y=y,
                                               state=state, index=index,
                                               is_request=is_request,
                                               is_goto=is_goto,
                                               ))
        f.write(gen_composite(screen='GRD_STATES_TARGET.adl',
                              macros={'STATE_N': index},
                              x=x+270, y=y,
                              width=25, height=25))

    lx = 17
    ly = button_top+rows*state_height+8
    if cols > 1:
        lx = margin + state_width + 7
        ly = title_height + margin + 28
    f.write(gen_composite(screen='GRD_STATES_LEGEND.adl',
                          x=lx, y=ly,
                          width=315, height=18))

def write_screen_spm(f):
    with open(get_path('GRD_SPM.adl')) as fs:
        f.write(fs.read())

def write_screen_usermsg(f):
    with open(get_path('GRD_USERMSG.adl')) as fs:
        f.write(fs.read())

def write_composite_status(f, x=0, y=0):
    f.write(gen_composite(screen='GRD_STATUSBAR.adl',
                          x=x, y=y, width=10, height=10))
